"""IDLE Extension Utilities."""

# Programmed by CoolCat467

from __future__ import annotations

# IDLE Extension Utilities
# Copyright (C) 2023-2025  CoolCat467
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__title__ = "extension-utils"
__author__ = "CoolCat467"
__license__ = "GNU General Public License Version 3"

import importlib
import sys
import time
import traceback
from contextlib import contextmanager
from functools import wraps
from idlelib import search, searchengine
from idlelib.config import idleConf
from os.path import abspath
from pathlib import Path
from tkinter import TclError, Text, Tk, messagebox
from typing import TYPE_CHECKING, ClassVar, Literal, NamedTuple, TypeVar

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Iterable, Sequence
    from idlelib.editor import EditorWindow
    from idlelib.format import FormatRegion
    from idlelib.iomenu import IOBinding
    from idlelib.pyshell import PyShellEditorWindow, PyShellFileList
    from idlelib.undo import UndoDelegator

    from typing_extensions import ParamSpec, Self

    PS = ParamSpec("PS")

T = TypeVar("T")

LOGS_PATH = Path(idleConf.userdir) / "logs"
TITLE: str = __title__


def set_title(title: str) -> None:
    """Set program title."""
    global TITLE
    TITLE = title


def get_required_config(
    values: dict[str, str],
    bind_defaults: dict[str, str | None],
    extension_title: str,
) -> str:
    """Get required configuration file data."""
    if __title__ == TITLE:
        set_title(extension_title)
    config = ""
    # Get configuration defaults
    settings = "\n".join(
        f"{key} = {default}" for key, default in values.items()
    )
    if settings:
        config += f"\n[{extension_title}]\n{settings}"
        if bind_defaults:
            config += "\n"
    # Get key bindings data
    settings = "\n".join(
        f"{event} = {key}"
        for event, key in bind_defaults.items()
        if key is not None
    )
    if settings:
        config += f"\n[{extension_title}_cfgBindings]\n{settings}"
    return config


def check_installed(
    extension: str,
    version: str,
    cls: type[BaseExtension] | None = None,
) -> bool:
    """Make sure extension installed. Return True if installed correctly."""
    # Get list of system extensions
    extensions = set(idleConf.defaultCfg["extensions"])

    # Do we have the user extend extension?
    has_user = "idleuserextend" in idleConf.GetExtensions(active_only=True)

    # If we don't, things get messy and we need to change the root config file
    ex_defaults = idleConf.defaultCfg["extensions"].file
    if has_user:
        # Otherwise, idleuserextend patches IDLE and we only need to modify
        # the user config file
        ex_defaults = idleConf.userCfg["extensions"].file
        extensions |= set(idleConf.userCfg["extensions"])

    if cls is None:
        # Import extension
        module = importlib.import_module(extension)

        # Get extension class
        if not hasattr(module, extension):
            print(
                f"ERROR: Somehow, {extension} was installed improperly, "
                f"no {extension} class found in module. Please report "
                "this on github.",
                file=sys.stderr,
            )
            return False

        cls = getattr(module, extension)
    # if not issubclass(cls, BaseExtension):
    #     raise ValueError(f"Expected BaseExtension subclass, got {cls!r}")

    # If this extension not in there,
    if extension not in extensions:
        # Get extension class keybinding defaults
        required_config = get_required_config(
            getattr(cls, "values", {}),
            getattr(cls, "bind_defaults", {}),
            extension,
        )

        # Tell user how to add it to system list.
        print(f"{extension} not in system registered extensions!")
        print(
            f"Please run the following command to add {extension} "
            + "to system extensions list.\n",
        )
        # Make sure line-breaks will go properly in terminal
        add_data = required_config.replace("\n", "\\n")
        # Tell them the command
        append = "| sudo tee -a"
        if has_user:
            append = ">>"
        print(f"echo -e '{add_data}' {append} {ex_defaults}\n")
    else:
        print(f"Configuration should be good! (v{version})")
        return True
    return False


def get_line_selection(line: int, length: int = 1) -> tuple[str, str]:
    """Get selection strings for given line(s)."""
    return f"{line}.0", f"{line + length}.0"


# Stolen from idlelib.searchengine
def get_line_col(index: str) -> tuple[int, int]:
    """Return (line, col) tuple of integers from {line}.{col} string."""
    line, col = map(int, index.split(".", 1))  # Fails on invalid index
    return line, col


# Stolen from idlelib.searchengine
def get_selected_text_indexes(text: Text) -> tuple[str, str]:
    """Return tuple of {line}.{col} indexes from selection or insert mark."""
    try:
        first = text.index("sel.first")
    except TclError:
        first = None
    try:
        last = text.index("sel.last")
    except TclError:
        last = None
    first = first or text.index("insert")
    last = last or first
    return first, last


def hide_hit(text: Text) -> None:
    """Remove `hit` tag from entire file."""
    text.tag_remove("hit", "1.0", "end")
    # text.update_idletasks()


def set_insert_and_move(text: Text, index: str) -> None:
    """Bring area into view.

    Moves `insert` mark and moves to insert mark.
    """
    text.mark_set("insert", index)
    text.see("insert")

    # Force update
    text.update_idletasks()


def highlight_region(text: Text, tag: str, first: str, last: str) -> None:
    """Add a given tag to the region of text between first and last indices."""
    if first == last:
        text.tag_add(tag, first)
    else:
        text.tag_add(tag, first, last)


def show_hit(text: Text, first: str, last: str, tag: str = "hit") -> None:
    """Highlight text between first and last indices.

    Indexes are formatted as `{line}.{col}` strings.

    Text is highlighted via the 'hit' tag and the marked
    section is brought into view.

    Does not clear previously set hit tags implicitly.

    Note that because of how IDLE works, selection tag ("sel")
    will override "hit" tag, so this function removes selection
    tags from the entire file. If you need this information,
    please use `get_selected_text_indexes` or something equivalent
    beforehand.
    """
    text.tag_remove("sel", "1.0", "end")
    highlight_region(text, tag, first, last)

    set_insert_and_move(text, first)


def get_whole_line(index: str, offset: int = 0) -> str:
    """Return index line plus offset at column zero."""
    line = get_line_col(index)[0]
    return f"{line + offset}.0"


def get_line_indent(text: str, char: str = " ") -> int:
    """Return line indent."""
    index = -1
    for index, cur_char in enumerate(text):
        if cur_char != char:
            return index
    return index + 1


def get_line_indent_handle_tabs(text: str) -> tuple[bool, int]:
    """Return if uses tabs and associated indentation level."""
    uses_tabs = text.startswith("\t")
    if uses_tabs:
        indent = get_line_indent(text, "\t")
    else:
        indent = get_line_indent(text)
    return uses_tabs, indent


def ensure_section_exists(section: str) -> bool:
    """Ensure section exists in user extensions configuration.

    Returns True if edited.
    """
    if section not in idleConf.GetSectionList("user", "extensions"):
        idleConf.userCfg["extensions"].AddSection(section)
        return True
    return False


def ensure_values_exist_in_section(
    section: str,
    values: dict[str, str | None] | dict[str, str],
) -> bool:
    """For each key in values, make sure key exists. Return if edited.

    If key does not exist and default value is not None, create and set
    to value.
    """
    need_save = False
    for key, default in values.items():
        value = idleConf.GetOption(
            "extensions",
            section,
            key,
            warn_on_default=False,
        )
        if value is None and default is not None:
            idleConf.SetOption("extensions", section, key, default)
            need_save = True
    return need_save


def ask_save_dialog(parent: Text) -> bool:
    """Ask to save dialog. Return if ok to save.

    Stolen from idlelib.runscript.ScriptBinding.
    """
    msg = "Source Must Be Saved\n" + 5 * " " + "OK to Save?"
    confirm: bool = messagebox.askokcancel(
        title="Save Before Run or Check",
        message=msg,
        default=messagebox.OK,
        parent=parent,
    )
    return confirm


def get_search_engine_params(
    engine: searchengine.SearchEngine,
) -> dict[str, str | bool]:
    """Get current search engine parameters."""
    return {
        name: getattr(engine, f"{name}var").get()
        for name in ("pat", "re", "case", "word", "wrap", "back")
    }


def set_search_engine_params(
    engine: searchengine.SearchEngine,
    data: dict[str, str | bool],
) -> None:
    """Get current search engine parameters."""
    for name in ("pat", "re", "case", "word", "wrap", "back"):
        if name in data:
            getattr(engine, f"{name}var").set(data[name])


@contextmanager
def search_engine_block(
    engine: searchengine.SearchEngine,
) -> Generator[None, None, None]:
    """Search engine modification context manager."""
    global_search_params = get_search_engine_params(engine)
    try:
        yield None
    finally:
        set_search_engine_params(engine, global_search_params)


@contextmanager
def undo_block(undo: UndoDelegator) -> Generator[None, None, None]:
    """Undo block context manager."""
    undo.undo_block_start()
    try:
        yield None
    finally:
        undo.undo_block_stop()


@contextmanager
def temporary_overwrite(
    object_: object,
    attribute: str,
    value: object,
) -> Generator[None, None, None]:
    """Temporarily overwrite object_.attribute with value, restore on exit."""
    if not hasattr(object_, attribute):
        yield None
    else:
        original = getattr(object_, attribute)
        setattr(object_, attribute, value)
        try:
            yield None
        finally:
            setattr(object_, attribute, original)


def extension_log(content: str) -> None:
    """Log content to extension log file."""
    if not LOGS_PATH.exists():
        LOGS_PATH.mkdir(exist_ok=True)
    log_file = LOGS_PATH / f"{TITLE}.log"
    with log_file.open("a", encoding="utf-8") as fp:
        format_time = time.strftime("[%Y-%m-%d %H:%M:%S] ")
        for line in content.splitlines(keepends=True):
            fp.write(f"{format_time}{line}")
        if not line.endswith("\n"):
            fp.write("\n")


def extension_log_exception(exc: BaseException, print_: bool = True) -> None:
    """Log exception to extension log."""
    exception_text = "".join(traceback.format_exception(exc))
    extension_log(exception_text)
    if print_:
        print(exception_text, file=sys.stderr)


def log_exceptions(function: Callable[PS, T]) -> Callable[PS, T]:
    """Log any exceptions raised."""

    @wraps(function)
    def wrapper(*args: PS.args, **kwargs: PS.kwargs) -> T:
        """Catch Exceptions, log them to log file, and re-raise."""
        try:
            return function(*args, **kwargs)
        except Exception as exc:
            extension_log_exception(exc)
            raise

    return wrapper


class Comment(NamedTuple):
    """Represents one comment."""

    file: str
    line: int
    contents: str
    line_end: int | None = None
    column: int = 0
    column_end: int | None = None

    def get_full_span(self) -> tuple[int, int, int, int]:
        """Return full span as tuple of (line, col, end_line, end_col)."""
        return (
            self.line,
            self.column,
            self.line_end if self.line_end is not None else self.line,
            self.column_end if self.column_end is not None else self.column,
        )

    def replace_content(self, contents: str) -> Self:
        """Return comment with same data except contents."""
        return self._replace(contents=contents)


def int_default(text: str, default: int = 0) -> int:
    """Return text as int or default if there is a ValueError."""
    try:
        return int(text)
    except ValueError:
        return default


class FilePosition(NamedTuple):
    """File Position."""

    path: str
    line: int
    col: int
    line_end: int
    col_end: int

    def is_range(self) -> bool:
        """Return True if file position covers a range."""
        return self.line != self.line_end or self.col != self.col_end

    def as_select(self) -> tuple[str, str]:
        """Return text selection region index strings."""
        return f"{self.line}.{self.col}", f"{self.line_end}.{self.col_end}"

    def delta_column(self, delta: int = -1) -> Self:
        """Return position but with delta added to column."""
        return self._replace(col=self.col + delta)

    @classmethod
    def parse(cls, file_position: str) -> Self:
        """Parse file position string."""
        line = 0
        line_end = 0
        col = 0
        col_end = 0

        windows_drive_letter = ""
        if sys.platform == "win32":
            windows_drive_letter, file_position = file_position.split(":", 1)
            windows_drive_letter += ":"
        position = file_position.split(":", 5)

        filename = position[0]
        if len(position) > 1:
            line = int_default(position[1])
            line_end = line
        if len(position) > 2:
            col = int_default(position[2])
            col_end = col
        if len(position) > 4:
            line_end = int_default(position[3], line_end)
            col_end = int_default(position[4], col_end)

        # If line end is before beginning, swap.
        if line_end < line:
            line, line_end = line_end, line
            col, col_end = col_end, col

        return cls(
            path=f"{windows_drive_letter}{filename}",
            line=line,
            col=col,
            line_end=line_end,
            col_end=col_end,
        )


class BaseExtension:
    """Base extension class."""

    __slots__ = (
        "comment_prefix",
        "editwin",
        "files",
        "flist",
        "formatter",
        "text",
        "undo",
    )

    # Extend the file and format menus.
    menudefs: ClassVar[
        Sequence[tuple[str, Sequence[tuple[str, str] | None]]]
    ] = ()

    # Default values for configuration file
    values: ClassVar[dict[str, str]] = {
        "enable": "True",
        "enable_editor": "True",
        "enable_shell": "False",
    }

    # Default key binds for configuration file
    bind_defaults: ClassVar[dict[str, str | None]] = {}

    def __init__(
        self,
        editwin: PyShellEditorWindow,
        *,
        comment_prefix: str | None = None,
    ) -> None:
        """Initialize this extension."""
        self.editwin: PyShellEditorWindow = editwin
        self.text: Text = editwin.text
        self.undo: UndoDelegator = editwin.undo
        self.formatter: FormatRegion = editwin.fregion
        self.files: IOBinding = editwin.io
        self.flist: PyShellFileList = editwin.flist

        if comment_prefix is None:
            comment_prefix = f"{self.__class__.__name__}"
        self.comment_prefix = f"# {comment_prefix}: "

        self.bind_non_keyboard(self.bind_defaults)

    def __repr__(self) -> str:
        """Return representation of self."""
        return f"{self.__class__.__name__}({self.editwin!r})"

    def bind_non_keyboard(self, bind_defaults: dict[str, str | None]) -> None:
        """Bind non-keyboard triggered events.

        IDLE only binds keyboard events automatically.
        """
        for bind_name, key in bind_defaults.items():
            if key is not None:
                continue
            bind_func_name = bind_name.replace("-", "_") + "_event"
            if not hasattr(self, bind_func_name):
                raise ValueError(f"Missing function {bind_func_name}")
            bind_func = getattr(self, bind_func_name)
            if not callable(bind_func):
                raise ValueError(f"{bind_func_name} should be callable")
            self.text.bind(f"<<{bind_name}>>", bind_func)

    def get_rightclick_menu_labels(self) -> list[str | None]:
        """Return rightclick menu labels."""
        return [entry[0] for entry in self.editwin.rmenu_specs]

    def register_rightclick_menu_entry(
        self,
        label: str,
        event_name: str,
        verify_function: Callable[[], bool] | None = None,
    ) -> None:
        """Register a rightclick menu entry.

        Arguments:
        label: str
            Text to display.
        event_name: str
            Event to raise when entry is selected.
        verify_function: Callable[[], bool] | None
            Optional function to determine if this entry should
            be enabled or not. Returns True if should be enabled.

        Skips adding if label already in right click menu labels.

        """
        if label in self.get_rightclick_menu_labels():
            return

        entry: tuple[str, str, str | None]
        if verify_function is None:
            entry = (label, event_name, None)
        else:
            # Technically can be "normal", "disabled", or "active",
            # where "active" sets it to like it's being hovered
            # but we will say no.
            # Normal is standard unlit, disabled is disabled.
            @wraps(verify_function)
            def verify_state_return_wrap() -> Literal["normal" | "disabled"]:
                return "normal" if verify_function() else "disabled"

            # wacky thing, idlelib.editor.right_menu_event does
            # `state = getattr(self, verify_state)()`
            # so needs to be an attribute on editor object.
            attr_name = f"_rmenu_verify_function_{label.lower()}"
            setattr(self.editwin, attr_name, verify_state_return_wrap)

            entry = (label, event_name, attr_name)
        self.editwin.rmenu_specs.append(entry)

    def register_rightclick_menu_entries(
        self,
        entries: Iterable[tuple[str, str, Callable[[], bool] | None]],
    ) -> None:
        """Register multiple rightclick menu entries.

        Entries are a series of (label, event_name, verify_function) tuples.

        See register_rightclick_menu_entry for more information.
        """
        for label, event_name, verify_function in entries:
            self.register_rightclick_menu_entry(
                label,
                event_name,
                verify_function,
            )

    @classmethod
    def ensure_bindings_exist(cls) -> bool:
        """Ensure key bindings exist in user extensions configuration.

        Return True if need to save.
        """
        if not cls.bind_defaults:
            return False

        need_save = False
        section = f"{cls.__name__}_cfgBindings"
        if ensure_section_exists(section):
            need_save = True
        if ensure_values_exist_in_section(section, cls.bind_defaults):
            need_save = True
        return need_save

    @classmethod
    def ensure_config_exists(cls) -> bool:
        """Ensure required configuration exists for this extension.

        Return True if need to save.
        """
        need_save = False
        if ensure_section_exists(cls.__name__):
            need_save = True
        if ensure_values_exist_in_section(cls.__name__, cls.values):
            need_save = True
        return need_save

    @classmethod
    def reload(cls) -> None:
        """Load class variables from configuration."""
        # Ensure file default values exist so they appear in settings menu
        save = cls.ensure_config_exists()
        if cls.ensure_bindings_exist() or save:
            idleConf.SaveUserCfgFiles()

        # Reload configuration file
        idleConf.LoadCfgFiles()

        # For all possible configuration values
        for key, default in cls.values.items():
            # Set attribute of key name to key value from configuration file
            if key not in {"enable", "enable_editor", "enable_shell"}:
                value = idleConf.GetOption(
                    "extensions",
                    cls.__name__,
                    key,
                    default=default,
                )
                setattr(cls, key, value)

    def get_tabwidth_indent_spaces(self) -> str:
        """Return tabwidth indent as spaces."""
        return " " * self.editwin.get_tk_tabwidth()

    def get_line(
        self,
        line: int,
        text_win: Text | None = None,
    ) -> str:
        """Get the characters from the given line in currently open file."""
        if text_win is None:
            text_win = self.text
        chars: str = text_win.get(*get_line_selection(line))
        return chars

    def get_line_replace_tabs(
        self,
        line: int,
        text_win: Text | None = None,
    ) -> tuple[bool, str]:
        """Return if line uses tabs and line using spaces."""
        chars = self.get_line(line, text_win)
        if chars.startswith("\t"):
            return True, chars.replace("\t", self.get_tabwidth_indent_spaces())
        return False, chars

    def reinstate_line_tabs(self, line: str) -> str:
        """Return line with leading indent replaced with tabs."""
        indent = get_line_indent(line)
        indent_text = line[:indent]
        post_indent = line[indent:]
        return (
            indent_text.replace(self.get_tabwidth_indent_spaces(), "\t")
            + post_indent
        )

    def reinstate_char_tabs(self, chars: str) -> str:
        """Return potentially multiline string with indentation replaced with tabs."""
        lines = chars.splitlines(keepends=True)
        return "".join(map(self.reinstate_line_tabs, lines))

    def get_comment_line(self, indent: int, content: str) -> str:
        """Return comment line given indent and content."""
        strindent = " " * indent
        return f"{strindent}{self.comment_prefix}{content}"

    def comment_exists(
        self,
        line: int,
        comment: str,
        text_win: Text | None = None,
    ) -> bool:
        """Return True if comment for message already exists on line."""
        return self.get_comment_line(0, comment) in self.get_line(
            line - 1,
            text_win=text_win,
        )

    def add_comment(
        self,
        comment: Comment,
        max_exist_up: int = 0,
    ) -> bool:
        """Return True if added new comment, False if already exists.

        Arguments:
        ---------
            max_exist_up: Max distance upwards to look for comment to already exist.

        Does not use an undo block, please use one yourself.

        """
        # Get line and message from output
        file = comment.file
        line = comment.line
        msg = comment.contents

        editwin: EditorWindow = self.editwin

        open_file: str | None = self.files.filename
        if open_file is None or abspath(open_file) != file:
            opened = self.flist.open(file)
            if opened is None:
                return False
            editwin = opened

        # If there is already a comment from us there, ignore that line.
        # +1-1 is so at least up by 1 is checked, range(0) = []
        for i in range(max_exist_up + 1):
            if self.comment_exists(line - (i - 1), msg, editwin.text):
                return False

        # Get line checker is talking about
        uses_tabs, chars = self.get_line_replace_tabs(line, editwin.text)

        # Figure out line indent
        indent = get_line_indent(chars)

        # Add comment line
        chars = self.get_comment_line(indent, msg) + "\n" + chars
        if uses_tabs:
            chars = self.reinstate_char_tabs(chars)

        # Save changes
        start, end = get_line_selection(line)
        editwin.text.delete(start, end)
        editwin.text.insert(start, chars, ())
        return True

    def get_pointers(self, comments: list[Comment]) -> Comment | None:
        """Return comment pointing to multiple comments all on the same line.

        If none of the comment pointers are going to be visible
        with the comment prefix, returns None.

        Does not handle comments that span multiple lines, assumes
        comments are all comment.line

        Messages must all be on the same line and be in the same file,
        otherwise ValueError is raised.
        """
        line = comments[0].line
        file = comments[0].file

        # Figure out next line intent
        next_line_text = self.get_line(line + 1)
        _uses_tabs, indent = get_line_indent_handle_tabs(next_line_text)

        lastcol = len(self.get_comment_line(indent, ""))

        columns: set[int] = set()

        for comment in comments:
            if comment.line != line:
                raise ValueError(f"Comment `{comment}` not on line `{line}`")
            if comment.file != file:
                raise ValueError(f"Comment `{comment}` not in file `{file}`")
            if comment.column_end is None:
                end = comment.column
            else:
                end = comment.column_end
            for col in range(comment.column, end + 1):
                columns.add(col)

        new_line = ""
        for col in sorted(columns):
            spaces = (col - lastcol) - 1
            if spaces < 0:
                continue
            new_line += " " * spaces + "^"
            lastcol = col

        if not new_line.strip():
            return None

        return Comment(file=file, line=line + 1, contents=new_line)

    def add_comments(
        self,
        comments: Sequence[Comment],
    ) -> dict[str, list[int]]:
        """Add comments to file(s). Ignores comments that already exist.

        Return dict of per file a list of lines were a comment was added.

        Changes are wrapped in an undo block.
        """
        file_comments: dict[str, list[int]] = {}

        with undo_block(self.undo):
            total = len(comments)
            for comment in reversed(comments):
                if self.add_comment(comment, total):
                    file_comments.setdefault(comment.file, [])
                    file_comments[comment.file].append(comment.line)
        return file_comments

    def add_comment_block(
        self,
        file: str,
        start_line: int,
        lines: Sequence[str],
    ) -> list[int]:
        """Add lines to file, in order as they appear top to bottom.

        Returns list of lines were a comment was added.

        Changes are wrapped in an undo block.
        """
        if not lines:
            return []
        file_comments = self.add_comments(
            [
                Comment(
                    file=file,
                    line=start_line,
                    contents=line,
                )
                for line in lines
            ],
        )
        return file_comments.get(file, [])

    def remove_selected_extension_comments(self) -> bool:
        """Remove selected extension comments. Return if removed any comments.

        Changes are wrapped in an undo block.
        """
        # Get selected region lines
        head, _tail, chars, lines = self.formatter.get_region()
        region_start, _col = get_line_col(head)

        edited = False
        with undo_block(self.undo):
            for index, line_text in reversed(tuple(enumerate(lines))):
                # If after indent there is mypy comment
                if line_text.lstrip().startswith(self.comment_prefix):
                    # If so, remove line
                    self.text.delete(
                        *get_line_selection(index + region_start),
                    )
                    edited = True
        if not edited:
            # Make bell sound so user knows this ran even though
            # nothing happened.
            self.text.bell()
        return edited

    def remove_all_extension_comments(self) -> str:
        """Remove all extension comments.

        Changes are wrapped in an undo block.
        """
        eof_idx = self.text.index("end")
        chars = self.text.get("0.0", eof_idx)

        lines = chars.splitlines()

        edited = False
        with undo_block(self.undo):
            for index, line_text in reversed(tuple(enumerate(lines))):
                # If after indent there is mypy comment
                if line_text.lstrip().startswith(self.comment_prefix):
                    # If so, remove line
                    self.text.delete(*get_line_selection(index))
                    edited = True
        if not edited:
            # Make bell sound so user knows this ran even though
            # nothing happened.
            self.text.bell()
        return "break"

    def find_next_extension_comment(self, search_wrap: bool = True) -> bool:
        """Find next extension comment by hacking the search dialog engine.

        Return True if the search was successful and False otherwise.
        """
        root: Tk = self.editwin.root

        # Get search engine singleton from root
        engine: searchengine.SearchEngine = searchengine.get(root)

        # With search engine parameter restore block
        with search_engine_block(engine):
            # Set search pattern to comment starter
            set_search_engine_params(
                engine,
                {
                    "pat": f"^\\s*{self.comment_prefix}",
                    "re": True,
                    "case": True,
                    "word": False,
                    "wrap": search_wrap,
                    "back": False,
                },
            )

            # Find current pattern
            found = search.find_again(self.text)
            assert isinstance(found, bool)
            return found



"""Annotate - Annotate functions given signature."""

# Extensively modified by CoolCat467
# from https://github.com/dropbox/pyannotate/blob/master/pyannotate_tools/annotations/parse.py

from __future__ import annotations

# Pyannotate is licensed under the terms of the Apache License, Version 2.0,
# reproduced below.
#   Copyright (c) 2017 Dropbox, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

__title__ = "Annotate"
__author__ = "CoolCat467"
__license__ = "Apache License, Version 2.0"


import re
from tokenize import TokenInfo, generate_tokens, tok_name
from typing import TYPE_CHECKING, Any, Final, NamedTuple, NoReturn

from idlemypyextension.utils import extension_log, extension_log_exception

if TYPE_CHECKING:
    from collections.abc import Callable, Collection, Generator, Sequence

TYPING_LOWER: Final = {"List", "Set", "Type", "Dict", "Tuple", "Overload"}


def debug(message: str) -> None:
    """Print debug message."""
    # TODO: Censor username/user files
    content = f"[{__title__}] DEBUG: {message}"
    print(f"\n{content}")
    extension_log(content)


class ParseError(Exception):
    """Raised on any type comment parse error."""


class Token(NamedTuple):
    """Base token."""

    text: str | None = None


class LambdaBody(Token):
    """Lambda Body Text Token."""

    __slots__ = ()


class Name(Token):
    """Name Token."""

    __slots__ = ()


class FunctionName(Token):
    """Function name."""

    __slots__ = ()


class ArgumentName(Name):
    """Argument name."""

    __slots__ = ()


class Operator(Token):
    """Operator Token."""

    __slots__ = ()


class DottedName(Name):
    """An identifier token, such as 'List', 'int' or 'package.name'."""

    __slots__ = ()


class ArgumentDefault(DottedName):
    """Argument value."""

    __slots__ = ()


class Separator(Operator):
    """A separator or punctuator token such as '(', '[' or '->'."""

    __slots__ = ()


class EndSeparator(Separator):
    """End separator that does not need extra space."""

    __slots__ = ()


class Colin(Separator):
    """Colin ':'."""

    __slots__ = ()


class TypeDef(Colin):
    """Type Definition Start ':'."""

    __slots__ = ()


class DefaultDef(Separator):
    """Argument Default Definition Start '='."""

    __slots__ = ()


class Keyword(Name):
    """Keyword such as 'async', 'def'."""

    __slots__ = ()


class ReturnTypeDef(Separator):
    """Return type definition '->'."""

    __slots__ = ()


class Definition(Keyword):
    """Definition keyword."""

    __slots__ = ()


class EndDefinition(Colin):
    """End Definition Colin."""

    __slots__ = ()


class End(Token):
    """A token representing the end of a type comment."""

    __slots__ = ()

    def __init__(self) -> None:
        """Initialize superclass with no arguments."""
        super().__init__()


TOKENIZE_PATTERN: Final = re.compile(r"[-\w`]+(\s*(\.|:)\s*[-/\w]*)*")


def tokenize(txt: str) -> list[Token]:
    """Translate a type comment into a list of tokens."""
    original = txt
    txt = txt.replace("?", "")
    tokens: list[Token] = []
    while True:
        if not txt:
            tokens.append(End())
            return tokens
        if txt[0] in {" ", "\n"}:
            txt = txt[1:]
        elif txt[0] in "()[],*|":
            tokens.append(Separator(txt[0]))
            txt = txt[1:]
        elif txt[:2] == "->":
            tokens.append(Separator("->"))
            txt = txt[2:]
        elif txt[:3] == "...":
            tokens.append(DottedName("..."))
            txt = txt[3:]
        else:
            # match_ = re.match(r"[-\w`]+(\s*(\.|:)\s*[-/\w]*)*", txt)
            match_ = TOKENIZE_PATTERN.match(txt)
            if not match_:
                raise ParseError(f"Could not parse {txt!r} from {original!r}")
            fullname = match_.group(0)
            size = len(fullname)
            fullname = fullname.replace(" ", "")
            if "`" in fullname:
                fullname = fullname.split("`")[0]
            # pytz creates classes with the name of the timezone being used:
            # https://github.com/stub42/pytz/blob/f55399cddbef67c56db1b83e0939ecc1e276cf42/src/pytz/tzfile.py#L120-L123
            # This causes pyannotate to crash as it's invalid to have a class
            # name with a `/` in it (e.g. "pytz.tzfile.America/Los_Angeles")
            if fullname.startswith("pytz.tzfile."):
                fullname = "datetime.tzinfo"
            if "-" in fullname or "/" in fullname:
                debug(f"tokenize bad fullname: {fullname!r} -> Any")
                # Not a valid Python name; there are many places that
                # generate these, so we just substitute Any rather
                # than crashing.
                fullname = "Any"
            tokens.append(DottedName(fullname))
            txt = txt[size:]


def get_line_indent(text: str, char: str = " ") -> int:
    """Return line indent."""
    index = -1
    for index, cur_char in enumerate(text):
        if cur_char != char:
            return index
    return index + 1


def read_lambda(
    generator: Generator[TokenInfo, None, None],
) -> tuple[str, bool]:
    """Return lambda body text and if had over read error."""
    text = []

    brackets = 0  # Current number of brackets in use
    has_args = False

    for token in generator:
        string = token.string
        type_ = tok_name[token.type]
        # debug(f'{string = }')
        if type_ == "OP":
            if string in "([{":
                # We have traveled in an bracket level
                brackets += 1
            elif string in ")]}":
                # Left a bracket level
                brackets -= 1
            elif string == ":":
                has_args = True
                string += " "
            elif string == ",":
                string += " "
        text.append(string)
        if (
            token.end[1] < len(token.line)
            and token.line[token.end[1]] in {",", "\n"}
            and has_args
            and not brackets
        ):
            return ("".join(text), False)
        if (
            token.end[1] < len(token.line)
            and string == ")"
            and has_args
            and brackets <= 0
        ):
            debug("read_lambda negative bracket counter, over-read.")
            return ("".join(text)[:-1], True)
    raise ParseError("Reading lambda failed")


def read_fstring(
    starting_token: TokenInfo,
    generator: Generator[TokenInfo, None, None],
) -> str:
    """Return f-string text."""
    text = [starting_token.string]
    match_pair = starting_token.string[1:]

    for token in generator:
        string = token.string
        type_ = tok_name[token.type]
        # debug(f'{string = }')
        if type_ == "FSTRING_START":
            string = read_fstring(token, generator)
        text.append(string)
        if type_ == "FSTRING_END" and string == match_pair:
            return "".join(text)
    # I don't think it's possible for f-string to fail without tokenization
    # errors before we get to that point.
    raise ParseError("Reading f-string failed")  # pragma: nocover


def tokenize_definition(
    start_line: int,
    get_line: Callable[[int], str],
) -> tuple[list[Token], int]:  # pragma: nocover
    """Return list of Tokens and number of lines after start."""
    current_line_no = start_line

    def read_line() -> str:
        """Return next line."""
        nonlocal current_line_no
        value = get_line(current_line_no)
        # print(f'read_line: {value!r}')
        current_line_no += 1
        return value

    hasdef = False  # Do we have a definition token?
    defstart = False  # Has definition started?
    brackets = 0  # Current number of brackets in use
    typedef = 0  # Brackets in type definition (truthy if in type definition)
    default_arg = 0  # Brackets in default argument value
    indent = 0  # Current indent
    tokens: list[Token] = []

    # For each token tokenizer module finds
    token_generator = generate_tokens(read_line)
    for token in token_generator:
        # print(f'{tok_name[token.type]} {token.string!r}', end=' -> ')
        string = token.string
        if tok_name[token.type] == "NAME":
            if string == "async":  # Remember async
                tokens.append(Keyword(string))
            elif (
                string == "def"
            ):  # Error if more than one in case invalid start
                if hasdef:
                    raise ParseError(
                        "Did not expect second definition keyword",
                    )
                tokens.append(Definition(string))
                hasdef = True  # Have definition now
            elif not defstart:  # If not started definition, only function name
                tokens.append(FunctionName(string))
            elif typedef:  # If we are doing type definition, add DottedName
                # If last was also a DottedName and it ended with a dot,
                # instead add name text to previous
                if (
                    tokens
                    and isinstance(tokens[-1], DottedName)
                    and tokens[-1].text is not None
                    and tokens[-1].text.endswith(".")
                ):
                    previous = tokens.pop().text
                    assert previous is not None
                    tokens.append(DottedName(previous + string))
                else:
                    tokens.append(DottedName(string))
            elif default_arg and string == "lambda":
                tokens.append(ArgumentDefault(string))
                body, over_read_error = read_lambda(token_generator)
                tokens.append(LambdaBody(body))
                # If over read, handle left bracket level.
                # Copy-paste of end bracket level from below:
                if over_read_error:
                    # Left a bracket level
                    brackets -= 1
                    assert not typedef
                    assert default_arg
                    default_arg -= 1
                    tokens.append(EndSeparator(")"))
            elif default_arg:
                # If defining argument default, add ArgumentDefault
                if tokens and isinstance(tokens[-1], ArgumentDefault):
                    assert tokens[-1].text is not None
                    last = tokens[-1].text.rstrip()[-1]
                    previous = tokens.pop().text
                    assert previous is not None
                    if last in {"+", "-"}:
                        tokens.append(ArgumentDefault(f"{previous} {string}"))
                    else:
                        tokens.append(ArgumentDefault(previous + string))
                else:
                    tokens.append(ArgumentDefault(string))
            else:  # Otherwise is an argument name
                tokens.append(ArgumentName(string))
        elif tok_name[token.type] == "OP":
            if string in "([{":
                # We have traveled in an bracket level
                brackets += 1
                # If we have a definition and this is open parenthesis,
                # we are starting definition
                if not defstart and hasdef and string == "(":
                    defstart = True
                # If doing a type definition, we have typedef level
                if typedef:
                    typedef += 1
                # if default_arg:
                #     default_arg += 1
                tokens.append(EndSeparator(string))  # No spaces
            elif string in ")]}":
                # Left a bracket level
                brackets -= 1
                if typedef:  # Maybe left a type def level too
                    typedef -= 1
                if default_arg:
                    default_arg -= 1
                tokens.append(EndSeparator(string))
            elif string in {"*", "**", "/"} and not default_arg:
                tokens.append(EndSeparator(string))
            elif string == "|":
                tokens.append(Separator(string))
            elif string == ",":
                # If exactly one level, leaving type definition land
                if typedef == 1:
                    typedef = 0
                if default_arg == 1:
                    default_arg = 0
                tokens.append(Separator(string))  # Space on end
            elif string == "->":
                # We are defining a return type
                tokens.append(ReturnTypeDef(string))
                typedef = 1
            elif string == ":":
                if defstart and not brackets:
                    # Ran out of brackets so done defining function
                    tokens.append(EndDefinition(string))
                    typedef = 0
                    break
                tokens.append(TypeDef(string))
                typedef = 1
            elif string == "=":  # Defining a default definition
                tokens.append(DefaultDef(string))
                typedef = 0  # shouldn't be defining type but make sure
                default_arg = 1
            elif string in {"-", "+", "~"}:  # Unary operators
                # Add on to previous ArgumentDefault if it exists
                if tokens and isinstance(tokens[-1], ArgumentDefault):
                    previous = tokens.pop().text
                    assert previous is not None
                    if string in {"-", "+"}:
                        tokens.append(ArgumentDefault(f"{previous} {string}"))
                    else:
                        tokens.append(ArgumentDefault(previous + string))
                else:
                    tokens.append(ArgumentDefault(string))
            elif (
                string
                in {
                    "/",
                    "//",
                    "*",
                    "^",
                    "@",
                    "%",
                    "**",
                    ">>",
                    "<<",
                    "&",
                    "|",
                }
                and tokens
                and isinstance(tokens[-1], ArgumentDefault)
            ):  # All other math operators
                prev = tokens.pop().text
                assert prev is not None
                tokens.append(ArgumentDefault(f"{prev} {string} "))
            elif string == "." and (
                tokens and isinstance(tokens[-1], DottedName)
            ):
                previous = tokens.pop().text
                assert previous is not None
                tokens.append(DottedName(previous + string))
            elif string == "...":  # Ellipsis constant
                tokens.append(DottedName(string))
            else:  # pragma: no cover
                raise ParseError(f"Exhaustive list of OP failed: {string!r}")
        elif tok_name[token.type] in {"NL", "NEWLINE"}:
            # replace separator ends with end separators
            if tokens and isinstance(tokens[-1], Separator):
                tokens.append(EndSeparator(tokens.pop().text))
            tokens.append(EndSeparator(string))
            # we have a new indent level
            indent = get_line_indent(get_line(current_line_no))
            tokens.append(EndSeparator(" " * indent))
        elif tok_name[token.type] == "COMMENT":
            # Remember comments as separators
            tokens.append(EndSeparator(f"  {string}"))
        elif tok_name[token.type] in {"STRING", "NUMBER"}:
            # Only argument default, so add on to previous if exists
            if tokens and isinstance(tokens[-1], ArgumentDefault):
                previous = tokens.pop().text
                assert previous is not None
                tokens.append(ArgumentDefault(previous + string))
            else:
                tokens.append(ArgumentDefault(string))
        elif tok_name[token.type] == "INDENT":  # pragma: nocover
            tokens.append(EndSeparator(string))
        elif tok_name[token.type] == "FSTRING_START":
            tokens.append(
                ArgumentDefault(read_fstring(token, token_generator)),
            )
        elif tok_name[token.type] == "ENDMARKER":
            debug(f"{tokens = }")
            raise EOFError(
                "Found ENDMARKER token while reading function definition",
            )
        else:  # pragma: nocover
            debug(f"{token = }")
            raise ParseError(
                f"Unrecognized token type {tok_name[token.type]!r}",
            )
        # print(tokens[-1])
    # print(tokens[-1])
    # print()
    tokens.append(End())
    return tokens, current_line_no - start_line


def get_type_repr(name: str, ignore_modules: set[str] | None = None) -> str:
    """Return representation of name."""
    if ignore_modules is None:
        ignore_modules = set()
    for separator in ".:":
        if separator in name:
            module, text = name.split(separator, 1)
            if module in ("typing", "mypy_extensions"):
                if text in TYPING_LOWER:
                    return text.lower()
                return text
            if module in ignore_modules:
                return text
    return name


def get_typevalue_repr(
    typevalue: TypeValue,
    ignore_modules: set[str] | None = None,
) -> str:
    """Return representation of ClassType."""
    name = get_type_repr(typevalue.name, ignore_modules)
    if name in TYPING_LOWER:
        name = name.lower()
    args = []
    for arg in typevalue.args:
        args.append(get_typevalue_repr(arg, ignore_modules))
    if not args:
        if name:
            return name
        return "[]"
    if name == "Union":
        return " | ".join(args)
    if name == "__hacks_DictItem":
        return ": ".join(args)
    values = ", ".join(args)
    return f"{name}[{values}]"


class TypeValue:
    """A type value, potentially collection of multiple."""

    __slots__ = ("args", "name")

    def __init__(
        self,
        name: str,
        args: Sequence[TypeValue] | None = None,
    ) -> None:
        """Set up name and arguments."""
        self.name = name
        if args:
            self.args = tuple(args)
        else:
            self.args = ()

    def __repr__(self) -> str:
        """Return representation of self."""
        args = f", {self.args!r}" if self.args else ""
        return f"TypeValue({self.name!r}{args})"

    def __str__(self) -> str:
        """Return type value representation of self."""
        return get_typevalue_repr(self)

    def __eq__(self, rhs: object) -> bool:
        """Return if rhs is equal to self."""
        if isinstance(rhs, self.__class__):
            return self.name == rhs.name and self.args == rhs.args
        return super().__eq__(rhs)


def list_or(values: Collection[str]) -> str:
    """Return comma separated listing of values joined with ` or `."""
    if len(values) <= 2:
        return " or ".join(values)
    copy = list(values)
    copy[-1] = f"or {copy[-1]}"
    return ", ".join(copy)


class Parser:
    """Implementation of the type comment parser."""

    __slots__ = ("i", "tokens")

    def __init__(self, tokens: list[Token]) -> None:
        """Initialize with tokens list."""
        self.tokens = tokens
        self.i = 0

    def parse_type_list(self) -> list[TypeValue]:
        """Parse comma separated type list."""
        types = []
        while self.lookup() not in (")", "]"):
            type_ = self.parse_type()
            types.append(type_)
            string = self.lookup()
            if string == ",":
                self.expect(",")
                continue
            self.expect_option((")", "]"))
            self.back()
        return types

    def parse_type_dict_or_set(self) -> list[TypeValue]:
        """Parse dict or set type list."""
        types = []
        # 0 -> Unknown
        # 1 -> dict
        # 2 -> set
        which = 0
        while self.lookup() != "}":
            type_ = self.parse_type()
            if which == 0:
                which = 1 if self.lookup() == ":" else 2
            if which == 1:  # dict
                self.expect(":")
                types.append(
                    TypeValue("__hacks_DictItem", (type_, self.parse_type())),
                )
            if which == 2:  # set
                types.append(type_)
            if self.lookup() == ",":
                self.expect(",")
                continue
            self.expect("}")
            self.back()
        return types

    def parse_type_collection(self, start_collection: str) -> list[TypeValue]:
        """Parse default argument type list."""
        if start_collection in ("(", "["):
            return self.parse_type_list()
        if start_collection == "{":
            return self.parse_type_dict_or_set()
        return self.fail(  # pragma: nocover
            "Unhandled collection start: {start_collection!r}",
        )

    def parse_union_list(self) -> list[TypeValue]:
        """Parse | separated union list."""
        types = []
        while True:
            type_ = self.parse_single()
            types.append(type_)
            if self.lookup() != "|":
                return types
            self.expect("|")

    def read_dotted_name_with_spaces(self, add: str = "") -> DottedName:
        """Parse dotted name token for modules with spaces in the filename."""
        token = self.expect_type(DottedName)
        if token.text is None:
            self.fail("Expected token text to be string, got None")
        if self.count_unprocessed_tokens() > 0 and isinstance(
            self.peek(),
            DottedName,
        ):
            return self.read_dotted_name_with_spaces(f"{add}{token.text} ")
        return DottedName(f"{add}{token.text}")

    def parse_single(self) -> TypeValue:
        """Parse single, does not look for | unions. Do not use directly."""
        if self.lookup() == "[":
            # Callable[[...], ...]
            #          ^^^^^
            self.expect("[")
            args = self.parse_type_list()
            self.expect("]")
            return TypeValue("", args)
        # token = self.expect_type(DottedName)
        # if token.text is None:
        #     self.fail("Expected token text to be string, got None")
        token = self.read_dotted_name_with_spaces()
        assert token.text is not None
        if token.text == "Any":
            return TypeValue("Any")
        if token.text in {"mypy_extensions.NoReturn", "typing.NoReturn"}:
            return TypeValue("NoReturn")
        if token.text == "Tuple":
            self.expect("[")
            args = self.parse_type_list()
            self.expect("]")
            return TypeValue(token.text, args)
        if token.text == "Union":
            self.expect("[")
            items = self.parse_type_list()
            self.expect("]")
            if len(items) == 1:
                return items[0]
            if len(items) == 0:
                self.fail("No items in Union")
            return TypeValue(token.text, items)
        if token.text == "lambda":
            lambda_data = self.parse_lambda()
            return TypeValue(f"{token.text} {lambda_data}")
        if token.text == "Overload":
            self.expect("(")
            args = self.parse_type_list()
            self.expect(")")
            return TypeValue(token.text, args)
        if self.lookup() == "[":
            self.expect("[")
            args = self.parse_type_list()
            self.expect("]")
            if token.text == "Optional" and len(args) == 1:
                return TypeValue("Union", [args[0], TypeValue("None")])
            return TypeValue(token.text, args)
        return TypeValue(token.text)

    def parse_lambda(self) -> str:
        """Parse lambda expression."""
        body = self.expect_type(LambdaBody)
        assert body.text is not None
        return body.text

    def parse_type(self) -> TypeValue:
        """Parse type including | unions."""
        type_value = self.parse_single()
        if self.lookup() == "|":
            self.expect("|")
            args = [type_value, *self.parse_union_list()]
            return TypeValue("Union", args)
        return type_value

    def fail(self, error: str | None) -> NoReturn:
        """Raise parse error."""
        raise ParseError(error)

    def peek(self) -> Token:
        """Peek at next token."""
        if self.i >= len(self.tokens):
            self.fail("Ran out of tokens")
        return self.tokens[self.i]

    def next(self) -> Token:
        """Get next token."""
        token = self.peek()
        self.i += 1
        return token

    def expect_option(self, options: Collection[str]) -> str:
        """Expect next token text to be one of options. Return matching text."""
        got = self.next().text
        if got not in options:
            expected = list_or(tuple(map(repr, options)))
            self.fail(f"Expected {expected}, got {got!r}")
        assert got is not None
        return got

    def expect(self, text: str) -> None:
        """Expect next token text to be text."""
        got = self.next().text
        if got != text:
            self.fail(f"Expected {text!r}, got {got!r}")

    def expect_type(
        self,
        token_type: type[Token] | tuple[type[Token], ...],
    ) -> Token:
        """Expect next token to be instance of token_type. Return token."""
        token = self.next()
        if not isinstance(token, token_type):
            if isinstance(token_type, tuple):
                expect_str = list_or(
                    [f"{cls.__name__!r}" for cls in token_type],
                )
            else:
                expect_str = f"{token_type.__name__!r}"
            self.fail(f"Expected {expect_str}, got {token!r}")
        return token

    def lookup(self) -> str:
        """Peek at next token and return it's text."""
        value = self.peek().text
        if value is None:
            return "None"
        return value

    def back(self) -> None:
        """Go back one token."""
        self.i -= 1

    def rest_tokens(self) -> list[Token]:
        """Return all tokens not processed."""
        return self.tokens[self.i : len(self.tokens)]

    def count_unprocessed_tokens(self) -> int:
        """Return number of unprocessed tokens."""
        return len(self.tokens) - self.i


def normalize_path(path: str) -> str:
    """Make all paths unix paths."""
    return path.replace("\\", "/")


def get_annotation(
    annotation: dict[str, Any],
    get_line: Callable[[int], str],
) -> tuple[str, int]:
    """Return annotation and end line."""
    # debug(f"{annotation = }\n")

    # Get definition tokens
    try:
        def_tokens, line_count = tokenize_definition(
            annotation["line"],
            get_line,
        )
    except ParseError as exc:
        debug(f"Could not tokenize definition\n{annotation = }")
        extension_log_exception(exc)
        raise
    except EOFError as exc:
        debug(
            f"Reached End of File, expected end of definition\n{annotation = }",
        )
        extension_log_exception(exc)

        raise ParseError(
            "Reached End of File, expected end of definition",
        ) from exc

    # Get the argument and return tokens from signature
    signature = annotation["signature"]
    arg_tokens = [tokenize(arg) for arg in signature["arg_types"]]
    return_tokens = tokenize(signature["return_type"])

    # Figure out module names that need to be ignored
    filename = normalize_path(annotation.get("path", "")).rsplit("/", 1)[-1]
    module_name = filename.rsplit(".", 1)[0]
    ignore_modules: set[str] = set()
    if module_name:
        ignore_modules.add(module_name)

    # Find start of argument definition
    new_lines = ""
    parser = Parser(def_tokens)
    while True:
        token = parser.next()
        if isinstance(token, ArgumentName):
            parser.back()
            break
        if token.text == ")":
            parser.back()
            break
        assert token.text is not None, (
            "Unreachable, End token is the only null text token"
        )
        new_lines += token.text
        if isinstance(token, Keyword):
            new_lines += " "

    # Find out which arguments we have to skip (*, **, /)
    skip_args: set[int] = set()
    # debug(f'{parser.rest_tokens() = }')
    argument = 0
    argparser = Parser(parser.rest_tokens())
    while True:
        if isinstance(argparser.peek(), ReturnTypeDef | EndDefinition):
            break

        token = argparser.next()
        if isinstance(token, Separator):
            if token.text in {"/", "*"}:
                skip_args.add(argument)
                argument += 1
                # if argparser.lookup() == "*":
                #     argparser.expect("*")

        elif isinstance(token, ArgumentName):
            argument += 1
        elif isinstance(token, End):  # pragma: nocover
            raise ParseError(
                "Found End token during argument location handling",
            )
    # debug(f"{skip_args = } {len(arg_tokens) = } {argument = }")

    arg_place = len(arg_tokens) - argument
    if arg_place < 0:
        # Self or class argument does not require type so annotations are
        # not given
        skip_args.add(0)
        arg_place = -1
    # debug(f"{skip_args = }")

    # Handle arguments
    argument = 0
    while True:
        token = parser.next()
        # debug(f"{token = } ({argument = })")
        if isinstance(token, Separator):
            if isinstance(token, EndSeparator):
                assert token.text is not None
                new_lines += token.text
            else:
                new_lines += f"{token.text} "
            if token.text in {"*", "/"}:
                argument += 1
            if token.text == ")":
                break
        elif isinstance(token, ArgumentName):
            name = token.text
            type_text = ""
            if isinstance(parser.peek(), TypeDef):
                parser.expect(":")
                type_value = parser.parse_type()
                type_text = ": " + get_typevalue_repr(
                    type_value,
                    ignore_modules,
                )
            elif argument not in skip_args:
                type_value = Parser(arg_tokens[arg_place]).parse_type()
                type_text = ": " + get_typevalue_repr(
                    type_value,
                    ignore_modules,
                )
            if isinstance(parser.peek(), DefaultDef):
                parser.expect("=")
                type_text += " = "
                if isinstance(parser.peek(), EndSeparator):
                    start_collection = parser.next().text
                    assert start_collection is not None
                    type_text += start_collection
                    type_text += ", ".join(
                        map(
                            str,
                            parser.parse_type_collection(start_collection),
                        ),
                    )
                    type_text += f"{parser.next().text}"
                else:
                    type_text += str(parser.parse_type())
            # debug(f"{name}{type_text}")
            new_lines += f"{name}{type_text}"
            argument += 1
            arg_place += 1
        elif isinstance(token, End):  # pragma: no cover
            debug("Found End token during argument parsing")
            parser.back()
            break
    # debug(f'{type_text = }')

    # Handle the end
    if isinstance(parser.peek(), EndDefinition | End):
        parser.tokens = [
            ReturnTypeDef("->"),
            *return_tokens[:-1],
            *parser.rest_tokens(),
        ]
        parser.i = 0
    # debug(f'{parser.rest_tokens() = }')
    parser.expect("->")
    new_lines += " -> "
    ret_type = str(parser.parse_type())
    new_lines += ret_type
    while True:
        token = parser.next()
        if isinstance(token, End):
            break
        assert token.text is not None
        new_lines += token.text

    return new_lines, line_count


def run(annotation: dict[str, object]) -> None:  # pragma: nocover
    """Run test of module."""

    def get_line(line: int) -> str:
        path = annotation["path"]
        assert isinstance(path, str), "Path must be string"
        with open(path, encoding="utf-8") as files:
            for line_no, line_text in enumerate(files):
                if line_no == line - 1:
                    return line_text
            raise EOFError

    print(f"{get_annotation(annotation, get_line)[0]}")


if __name__ == "__main__":  # pragma: nocover
    print(f"{__title__}\nProgrammed by {__author__}.\n")
    # run(annotation)



"""Client for mypy daemon mode - Modified version of mypy.dmypy.client.

This manages a daemon process which keeps useful state in memory
rather than having to read it back from disk on each run.
"""

# Modified by CoolCat467
# Original at https://github.com/python/mypy/blob/master/mypy/dmypy/client.py
# Original retrieved November 24th 2022
# Last updated to match original on March 11th 2023

# Mypy (and mypyc) are licensed under the terms of the MIT license,
# reproduced below.
#
#    The MIT License
#
#    Copyright (c) 2012-2022 Jukka Lehtosalo and contributors
#    Copyright (c) 2015-2022 Dropbox, Inc.
#
#    Permission is hereby granted, free of charge, to any person obtaining a
#    copy of this software and associated documentation files (the "Software"),
#    to deal in the Software without restriction, including without limitation
#    the rights to use, copy, modify, merge, publish, distribute, sublicense,
#    and/or sell copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following conditions:
#
#    The above copyright notice and this permission notice shall be included in
#    all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#    DEALINGS IN THE SOFTWARE.

from __future__ import annotations

__title__ = "Mypy Daemon Client"
__license__ = "MIT"

import base64
import contextlib
import sys
from collections import ChainMap
from typing import TYPE_CHECKING, Literal, TypedDict, cast

import orjson

from idlemypyextension.moduleguard import guard_imports
from idlemypyextension.utils import extension_log, extension_log_exception

with guard_imports({"trio", "mypy"}):
    import trio
    from mypy.dmypy_os import alive as _alive, kill as _kill
    from mypy.dmypy_server import (
        Server as _Server,
        daemonize as _daemonize,
        process_start_options as _process_start_options,
    )
    from mypy.ipc import IPCClient as _IPCClient, IPCException as _IPCException
    from mypy.version import __version__

if TYPE_CHECKING:
    import io
    import os
    from collections.abc import Sequence

    from typing_extensions import NotRequired


# If should force to use base request system mypy implements
# If False and not on windows, will manually read unix sockets asynchronously.
FORCE_BASE_REQUEST: bool = False


def debug(message: str) -> None:
    """Print debug message."""
    # TODO: Censor username/user files
    content = f"[{__title__}] DEBUG: {message}"
    print(f"\n{content}")
    extension_log(content)


class Stats(TypedDict):
    """Statistics dictionary from dmypy response."""

    validate_meta_time: float
    files_parsed: int
    modules_parsed: int
    stubs_parsed: int
    parse_time: float
    find_module_time: float
    find_module_calls: int
    graph_size: int
    stubs_found: int
    graph_load_time: float
    fm_cache_size: int
    load_fg_deps_time: float
    sccs_left: int
    nodes_left: int
    cache_commit_time: float


class Response(TypedDict):
    """Response dictionary from dmypy."""

    platform: NotRequired[str]
    python_version: NotRequired[str]
    out: NotRequired[str]
    err: NotRequired[str]
    stdout: NotRequired[str]
    stderr: NotRequired[str]
    error: NotRequired[str]
    memory_psutil_missing: NotRequired[str]
    memory_rss_mib: NotRequired[float]
    memory_vms_mib: NotRequired[float]
    memory_maxrss_mib: NotRequired[float]
    restart: NotRequired[str]
    status: NotRequired[int]
    stats: NotRequired[Stats]
    final: NotRequired[bool]


class BadStatusError(Exception):
    """Exception raised when there is something wrong with the status file.

    For example:
    - No status file found
    - Status file malformed
    - Process whose process id is in the status file does not exist
    """


# Client-side infrastructure.


def str_maintain_none(obj: str | object | None) -> str | None:
    """Return string or None, maintaining None."""
    if obj is None:
        return None
    return str(obj)


async def _read_status(
    status_file: str | os.PathLike[str],
) -> dict[str, object]:
    """Read status file.

    Raise BadStatusError if the status file doesn't exist or contains
    invalid JSON or the JSON is not a dict.
    """
    path = trio.Path(status_file)
    if not await path.is_file():
        raise BadStatusError("No status file found")

    contents = await path.read_text(encoding="utf-8")
    try:
        data = orjson.loads(contents)
    except Exception as exc:
        extension_log_exception(exc)
        raise BadStatusError("Malformed status file (not JSON)") from exc
    if not isinstance(data, dict):
        raise BadStatusError("Invalid status file (not a dict)")
    return data


def _check_status(data: dict[str, object]) -> tuple[int, str]:
    """Check if the process is alive.

    Return (process id, connection_name) on success.

    Raise BadStatusError if something's wrong.
    """
    if "pid" not in data:
        raise BadStatusError("Invalid status file (no pid field)")
    pid = data["pid"]
    if not isinstance(pid, int):
        raise BadStatusError("pid field is not an int")
    if not _alive(pid):
        raise BadStatusError("Daemon has died")
    if "connection_name" not in data:
        raise BadStatusError("Invalid status file (no connection_name field)")
    connection_name = data["connection_name"]
    if not isinstance(connection_name, str):
        raise BadStatusError("connection_name field is not a string")
    return pid, connection_name


async def get_status(status_file: str | os.PathLike[str]) -> tuple[int, str]:
    """Read status file and check if the process is alive.

    Return (process id, connection_name) on success.

    Raise BadStatusError if something's wrong.
    """
    data = await _read_status(status_file)
    return _check_status(data)


def _read_request_response_json(request_response: str | bytes) -> Response:
    """Read request response json."""
    # debug(f'{request_response = }')
    try:
        data = orjson.loads(request_response)
    except Exception as exc:
        extension_log_exception(exc)
        return {"error": "Data received is not valid JSON"}
    if not isinstance(data, dict):
        return {"error": f"Data received is not a dict (got {type(data)})"}
    return cast("Response", data)


async def _request_win32(
    name: str,
    request_arguments: bytes,
    timeout: int | None = None,  # noqa: ASYNC109
) -> Response:
    """Request from daemon on windows."""

    async def _receive(
        async_connection: trio._file_io.AsyncIOWrapper[io.StringIO],
    ) -> Response:
        """Receive JSON data from a connection until EOF.

        Raise OSError if the data received is not valid JSON or if it is
        not a dict.
        """
        bdata: str = await async_connection.read()
        if not bdata:
            return {"error": "No data received, IPC socket connection closed."}
        return _read_request_response_json(bdata)

    try:
        all_responses: list[Response] = []
        with _IPCClient(name, timeout) as client:
            async_client = trio.wrap_file(cast("io.StringIO", client))
            await async_client.write(request_arguments.decode("utf-8"))

            final = False
            while not final:
                response = await _receive(async_client)
                final = bool(response.pop("final", False))
                all_responses.append(response)
                # debug(f"{response = }")
        # if len(all_responses) > 1:
        #    debug(f"request win32 {all_responses = }")
        return cast("Response", dict(ChainMap(*all_responses).items()))  # type: ignore[arg-type]
    except (OSError, _IPCException, ValueError) as err:
        return {"error": str(err)}


def find_frame_in_buffer(
    buffer: bytearray,
) -> tuple[bytearray, bytearray | None]:
    """Return a full frame from the bytes we have in the buffer.

    Returns (next frame keep data, complete frame | None)
    """
    space_pos = buffer.find(b" ")
    if space_pos == -1:
        # Incomplete frame
        return buffer, None
    # We have a full frame
    return buffer[space_pos + 1 :], buffer[:space_pos]


async def _request_linux(
    filename: str,
    request_arguments: bytes,
    timeout: float | None = None,  # noqa: ASYNC109
) -> Response:
    """Request from daemon on linux/unix."""
    buffer = bytearray()
    frame: bytearray | None = None
    all_responses: list[Response] = []
    async with await trio.open_unix_socket(filename) as connection:
        # Frame the data by urlencoding it and separating by space.
        request_frame = base64.encodebytes(request_arguments) + b" "
        await connection.send_all(request_frame)

        is_not_done = True
        while is_not_done:
            # Receive more data into the buffer.
            try:
                if timeout is None:
                    more = await connection.receive_some()
                else:
                    with trio.fail_after(timeout):
                        more = await connection.receive_some()
            except trio.TooSlowError:
                return {"error": "IPC socket connection timed out"}
            if not more:
                # Connection closed
                # Socket was empty and we didn't get any frame.
                return {
                    "error": "No data received, IPC socket connection closed.",
                }
            buffer.extend(more)

            buffer, frame = find_frame_in_buffer(buffer)
            if frame is None:
                continue
            # Frame is not None, we read a full frame
            response_text = base64.decodebytes(frame)
            response = _read_request_response_json(response_text)

            is_not_done = not bool(response.pop("final", False))
            all_responses.append(response)
            # debug(f"{response = }")
    # if len(all_responses) > 1:
    #    debug(f"request linux {all_responses = }")
    return cast("Response", dict(ChainMap(*all_responses).items()))  # type: ignore[arg-type]


REQUEST_LOCK = trio.Lock()


async def request(
    status_file: str | os.PathLike[str],
    command: str,
    *,
    timeout: int | None = None,  # noqa: ASYNC109
    **kwds: object,
) -> Response:
    """Send a request to the daemon.

    Return the JSON dict with the response.

    Raise BadStatusError if there is something wrong with the status file
    or if the process whose process id is in the status file has died.

    Return {'error': <message>} if an IPC operation or receive()
    raised OSError.  This covers cases such as connection refused or
    closed prematurely as well as invalid JSON received.

    `stdout` and `stderr` fields are apparently debug data.
    """
    args = dict(kwds)
    args["command"] = command
    # Tell the server whether this request was NOT initiated from a
    # human-facing terminal, so that it can format the type checking
    # output accordingly.
    args["is_tty"] = False
    args["terminal_width"] = 80
    request_arguments = orjson.dumps(args)
    _pid, name = await get_status(status_file)

    async with REQUEST_LOCK:
        if sys.platform == "win32" or FORCE_BASE_REQUEST:
            return await _request_win32(name, request_arguments, timeout)
        # Windows run thinks unreachable, everything else knows it is not
        return await _request_linux(  # type: ignore[unreachable,unused-ignore]
            name,
            request_arguments,
            timeout,
        )


async def is_running(status_file: str | os.PathLike[str]) -> bool:
    """Check if the server is running cleanly."""
    try:
        await get_status(status_file)
    except BadStatusError:
        return False
    return True


async def stop(status_file: str | os.PathLike[str]) -> Response:
    """Stop daemon via a 'stop' request."""
    return await request(status_file, "stop", timeout=5)


async def _wait_for_server(
    status_file: str | os.PathLike[str],
    timeout: float = 5.0,  # noqa: ASYNC109
) -> bool:
    """Wait until the server is up. Return False if timed out."""
    try:
        with trio.fail_after(timeout):
            while True:
                try:
                    data = await _read_status(status_file)
                except BadStatusError:
                    # If the file isn't there yet, retry later.
                    await trio.sleep(0.1)
                    continue
                break
    except trio.TooSlowError:
        return False
    # If the file's content is bogus or the process is dead, fail.
    try:
        _check_status(data)
    except BadStatusError:
        return False
    return True


async def _start_server(
    status_file: str | os.PathLike[str],
    *,
    flags: list[str],
    daemon_timeout: int | None = None,
    allow_sources: bool = False,
    log_file: str | os.PathLike[str] | None = None,
) -> bool:
    """Start the server and wait for it. Return False if error starting.

    daemon_timeout:
        Server shutdown timeout (in seconds)
    """
    start_options = _process_start_options(flags, allow_sources)
    if (
        _daemonize(
            start_options,
            str(status_file),
            timeout=daemon_timeout,
            log_file=str_maintain_none(log_file),
        )
        != 0
    ):
        return False
    return await _wait_for_server(status_file)


async def restart(
    status_file: str | os.PathLike[str],
    *,
    flags: list[str],
    daemon_timeout: int | None = 0,
    allow_sources: bool = False,
    log_file: str | os.PathLike[str] | None = None,
) -> bool:
    """Restart daemon (it may or may not be running; but not hanging).

    Returns False if error starting.

    daemon_timeout:
        Server shutdown timeout (in seconds)
    """
    # Bad or missing status file or dead process; good to start.
    with contextlib.suppress(BadStatusError):
        await stop(status_file)
    return await _start_server(
        status_file,
        flags=flags,
        daemon_timeout=daemon_timeout,
        allow_sources=allow_sources,
        log_file=log_file,
    )


# Action functions


async def start(
    status_file: str | os.PathLike[str],
    *,
    flags: list[str],
    daemon_timeout: int = 0,
    allow_sources: bool = False,
    log_file: str | os.PathLike[str] | None = None,
) -> bool:
    """Start daemon if not already running.

    Returns False if error starting / already running.

    daemon_timeout:
        Server shutdown timeout (in seconds)
    """
    if not await is_running(status_file):
        # Bad or missing status file or dead process; good to start.
        return await _start_server(
            status_file,
            flags=flags,
            daemon_timeout=daemon_timeout,
            allow_sources=allow_sources,
            log_file=log_file,
        )
    return False


async def status(
    status_file: str | os.PathLike[str],
    *,
    timeout: int = 5,  # noqa: ASYNC109
    fswatcher_dump_file: str | None = None,
) -> Response:
    """Ask daemon to return status."""
    return await request(
        status_file,
        "status",
        timeout=timeout,
        fswatcher_dump_file=fswatcher_dump_file,
    )


async def run(
    status_file: str | os.PathLike[str],
    *,
    flags: list[str],
    timeout: int | None = None,  # noqa: ASYNC109
    daemon_timeout: int = 0,
    log_file: str | os.PathLike[str] | None = None,
    export_types: bool = False,
) -> Response:
    """Do a check, starting (or restarting) the daemon as necessary.

    Restarts the daemon if the running daemon reports that it is
    required (due to a configuration change, for example).

    daemon_timeout:
        Server shutdown timeout (in seconds)
    """
    # Start if missing status file or dead process
    await start(
        status_file=status_file,
        flags=flags,
        daemon_timeout=daemon_timeout,
        allow_sources=True,
        log_file=log_file,
    )
    response = await request(
        status_file,
        "run",
        timeout=timeout,
        version=__version__,
        args=flags,
        export_types=export_types,
    )
    # If the daemon signals that a restart is necessary, do it
    if "restart" in response:
        debug(f'Restarting: "{response["restart"]}"')
        await restart(
            status_file,
            flags=flags,
            daemon_timeout=timeout,
            allow_sources=True,
            log_file=log_file,
        )
        return await request(
            status_file,
            "run",
            timeout=timeout,
            version=__version__,
            args=flags,
            export_types=export_types,
        )
    return response


async def kill(status_file: str | os.PathLike[str]) -> bool:
    """Kill daemon process with SIGKILL. Return True if killed."""
    pid = (await get_status(status_file))[0]
    try:
        _kill(pid)
    except OSError as ex:
        debug(f"Kill exception: {ex}")
        return False
    return True


async def check(
    status_file: str | os.PathLike[str],
    files: Sequence[str],
    *,
    timeout: int | None = None,  # noqa: ASYNC109
    export_types: bool = False,
) -> Response:
    """Ask the daemon to check a list of files."""
    return await request(
        status_file,
        "check",
        timeout=timeout,
        files=files,
        export_types=export_types,
    )


async def recheck(
    status_file: str | os.PathLike[str],
    export_types: bool,
    *,
    timeout: int | None = None,  # noqa: ASYNC109
    remove: list[str] | None = None,
    update: list[str] | None = None,
) -> Response:
    """Ask the daemon to recheck the previous list of files.

    If at least one of --remove or --update is given, the server will
    update the list of files to check accordingly and assume that any other
    files are unchanged.  If none of these flags are given, the server will
    call stat() on each file last checked to determine its status.

    Files given in --update ought to exist.  Files given in --remove need not
    exist; if they don't they will be ignored.
    The lists may be empty but oughtn't contain duplicates or overlap.

    NOTE: The list of files is lost when the daemon is restarted.
    """
    # if remove is not None or update is not None:
    return await request(
        status_file,
        "recheck",
        timeout=timeout,
        export_types=export_types,
        remove=remove,
        update=update,
    )
    # return await request(
    #     status_file,
    #     "recheck",
    #     timeout=timeout,
    #     export_types=export_types,
    # )


async def inspect(
    status_file: str | os.PathLike[str],
    location: str,  # line:col
    show: Literal["type", "attrs", "definition"] = "type",
    *,
    timeout: int | None = None,  # noqa: ASYNC109
    verbosity: int = 0,
    limit: int = 0,
    include_span: bool = False,
    include_kind: bool = False,
    include_object_attrs: bool = False,
    union_attrs: bool = False,
    force_reload: bool = False,
) -> Response:
    """Ask daemon to print the type of an expression.

    Locate and statically inspect expression(s)

    Location specified as
    path/to/file.py:line:column[:end_line:end_column]. If position is
    given (i.e. only line and column), this will return all enclosing
    expressions

    show:
        What kind of inspection to run, can be one of the following:
        [type, attrs, definition]

    verbosity:
        Increase verbosity of the type string representation

    limit:
        Return at most NUM innermost expressions (if position is given);
        0 means no limit

    include_span:
        Prepend each inspection result with the span of corresponding
        expression (e.g. 1:2:3:4:"int")

    include_kind:
        Prepend each inspection result with the kind of corresponding
        expression (e.g. NameExpr:"int")

    include_object_attrs:
        Include attributes of "object" in "attrs" inspection

    union_attrs:
        Include attributes valid for some of possible expression types
        (by default an intersection is returned)

    force_reload:
        Re-parse and re-type-check file before inspection (may be slow)
    """
    return await request(
        status_file,
        "inspect",
        timeout=timeout,
        show=show,
        location=location,
        verbosity=verbosity,
        limit=limit,
        include_span=include_span,
        include_kind=include_kind,
        include_object_attrs=include_object_attrs,
        union_attrs=union_attrs,
        force_reload=force_reload,
    )


async def suggest(
    status_file: str | os.PathLike[str],
    function: str,
    do_json: bool,
    *,
    timeout: int | None = None,  # noqa: ASYNC109
    callsites: bool = False,
    no_errors: bool = False,
    no_any: bool = False,
    flex_any: float | None = None,
    use_fixme: str | None = None,
    max_guesses: int | None = 64,
) -> Response:
    """Ask the daemon for a suggested signature.

    Suggest a signature or show call sites for a specific function

    function:
        Function specified as '[package.]module.[class.]function' or
        absolute path with line number at the end

    do_json:
        Produce json that pyannotate can use to apply a suggestion

    callsites:
        Find callsites instead of suggesting a type

    no_errors:
        Only produce suggestions that cause no errors

    no_any:
        Only produce suggestions that don't contain Any

    flex_any:
        Allow anys in types if they go above a certain score (scores are from 0-1)

    use_fixme:
        A dummy name to use instead of Any for types that can't be inferred

    max_guesses:
        Set the maximum number of types to try for a function (default 64)
    """
    return await request(
        status_file,
        "suggest",
        timeout=timeout,
        function=function,
        json=do_json,
        callsites=callsites,
        no_errors=no_errors,
        no_any=no_any,
        flex_any=flex_any,
        use_fixme=use_fixme,
        max_guesses=max_guesses,
    )


async def hang(
    status_file: str | os.PathLike[str],
    *,
    timeout: int = 1,  # noqa: ASYNC109
) -> Response:
    """Hang for 100 seconds, as a debug hack."""
    if not isinstance(timeout, int):
        raise ValueError("Timeout must be an integer!")
    return await request(status_file, "hang", timeout=timeout)


def do_daemon(
    status_file: str | os.PathLike[str],
    flags: list[str],
    daemon_timeout: int | None = None,
) -> None:
    """Serve requests in the foreground.

    daemon_timeout:
        Server shutdown timeout (in seconds)
    """
    options = _process_start_options(flags, allow_sources=False)
    _Server(options, str(status_file), timeout=daemon_timeout).serve()



"""Type Check IDLE Extension."""

# Programmed by CoolCat467

from __future__ import annotations

# IDLE Mypy daemon integration extension
# Copyright (C) 2023-2025  CoolCat467
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__title__ = "extension"
__author__ = "CoolCat467"
__license__ = "GNU General Public License Version 3"

import contextlib
import math
import os
import re
import sys
import traceback
from functools import partial, wraps
from idlelib.config import idleConf
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Final, Literal

import orjson

from idlemypyextension import annotate, client, tktrio, utils

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from idlelib.pyshell import PyShellEditorWindow
    from tkinter import Event, Misc

DAEMON_TIMEOUT_MIN: Final = 5
ACTION_TIMEOUT_MIN: Final = 5
UNKNOWN_FILE: Final = "<unknown file>"
COULD_NOT_START_ERROR = client.Response(
    {"out": "", "err": "Error: Could not start mypy daemon"},
)


def debug(message: object) -> None:
    """Print debug message."""
    # Censor username/user files
    as_str = str(message)

    ##username = os.getlogin()
    ##home_directory = os.path.expanduser("~")
    ##
    ##as_str = as_str.replace(home_directory, "~")
    ##as_str = as_str.replace(username, "<username>")

    print(f"\n[{__title__}] DEBUG: {as_str}")


MYPY_ERROR_TYPE: Final = re.compile(r"  \[([a-z\-]+)\]\s*$")


def parse_comments(
    mypy_output: str,
    default_file: str = "<unknown file>",
    default_line: int = 0,
) -> dict[str, list[utils.Comment]]:
    """Parse mypy output, return mapping of filenames to lists of comments."""
    files: dict[str, list[utils.Comment]] = {}
    for output_line in mypy_output.splitlines():
        if not output_line.strip():
            continue
        filename = default_file
        line = default_line
        line_end = default_line
        col = 0
        col_end = 0
        msg_type = "unrecognized"

        if output_line.count(": ") < 2:
            text = output_line
        else:
            where, msg_type, text = output_line.split(": ", 2)

            windows_drive_letter = ""
            if sys.platform == "win32":
                windows_drive_letter, where = where.split(":", 1)
                windows_drive_letter += ":"
            position = where.rsplit(":", 4)

            filename = f"{windows_drive_letter}{position[0]}"
            colon_count = len(position)
            if colon_count > 1:
                line = int(position[1])
                line_end = line
            if colon_count > 2:
                col = int(position[2])
                col_end = col
            if colon_count > 4:
                line_end = int(position[3])
                # if line_end == line:
                col_end = int(position[4])
                # else:
                #    line_end = line
        comment_type = MYPY_ERROR_TYPE.search(text)
        if comment_type is not None:
            text = text[: comment_type.start()]
            msg_type = f"{comment_type.group(1)} {msg_type}"

        comment = utils.Comment(
            file=filename,
            line=line,
            contents=f"{msg_type}: {text}",
            column=col,
            line_end=line_end,
            column_end=col_end,
        )

        files.setdefault(filename, [])
        files[filename].append(comment)
    return files


def parse_type_inspect(
    mypy_output: str,
    file: str,
    default_line: int = 0,
) -> list[utils.Comment]:
    """Parse mypy inspect output as a list of comments."""
    comments = []
    for output_line in mypy_output.splitlines():
        if " -> " not in output_line:
            comments.append(utils.Comment(file, default_line, output_line))
            continue
        span, content = output_line.split(" -> ", 1)
        line, column, line_end, column_end = map(int, span.split(":", 3))
        content = content.removeprefix('"').removesuffix('"')
        comments.append(
            utils.Comment(file, line, content, line_end, column, column_end),
        )
    return comments


# Important weird: If event handler function returns 'break',
# then it prevents other bindings of same event type from running.
# If returns None, normal and others are also run.


class idlemypyextension(utils.BaseExtension):  # noqa: N801
    """Add comments from mypy to an open program."""

    __slots__ = ("triorun",)
    # Extend the file and format menus.
    menudefs: ClassVar[
        Sequence[tuple[str, Sequence[tuple[str, str] | None]]]
    ] = (
        (
            "edit",
            (
                None,
                ("_Type Check File", "<<type-check>>"),
                ("Find Next Type Comment", "<<find-next-type-comment>>"),
            ),
        ),
        (
            "format",
            (
                ("Suggest Signature", "<<suggest-signature>>"),
                ("Remove Type Comments", "<<remove-type-comments>>"),
            ),
        ),
        (
            "run",
            (("Shutdown dmypy daemon", "<<shutdown-dmypy-daemon>>"),),
        ),
    )
    # Default values for configuration file
    values: ClassVar[dict[str, str]] = {
        "enable": "True",
        "enable_editor": "True",
        "enable_shell": "False",
        "daemon_flags": "None",
        "search_wrap": "True",
        "suggest_replace": "False",
        "timeout_mins": "30",
        "action_max_sec": "None",
        "should_restart_always": "False",
    }
    # Default key binds for configuration file
    bind_defaults: ClassVar[dict[str, str | None]] = {
        "type-check": "<Alt-Key-t>",
        "suggest-signature": "<Alt-Key-s>",
        "remove-type-comments": "<Alt-Shift-Key-T>",
        "find-next-type-comment": "<Alt-Key-g>",
        "shutdown-dmypy-daemon": None,
        "dmypy-inspect-type": None,
        "dmypy-goto-definition": None,
        "dmypy-inspect-attrs": None,
    }

    # Overwritten in reload
    daemon_flags = "None"
    search_wrap = "True"
    suggest_replace = "False"
    timeout_mins = "30"
    action_max_sec = "None"
    should_restart_always = "False"

    # Class attributes
    idlerc_folder = Path(idleConf.userdir).expanduser().absolute()
    mypy_folder = idlerc_folder / "mypy"
    status_file = mypy_folder / "dmypy.json"
    log_file = mypy_folder / "log.txt"

    def __init__(self, editwin: PyShellEditorWindow) -> None:
        """Initialize the settings for this extension."""
        super().__init__(editwin, comment_prefix="types")

        if not self.mypy_folder.exists():
            self.mypy_folder.mkdir(parents=True)

        self.triorun = tktrio.TkTrioRunner(
            self.editwin.top,
            # self.flist,
            restore_close=self.editwin.close,
        )

        # Seems to be important to register after other
        # initialization tasks complete, otherwise sometimes
        # rightclick menu will be very laggy to open.
        self.text.after_idle(self.register_rightclick_items)

    def register_rightclick_items(self) -> None:
        """Register right click menu entries."""
        self.register_rightclick_menu_entries(
            (
                (
                    "Inspect Type",
                    "<<dmypy-inspect-type>>",
                    None,
                ),
                (
                    "Goto Definition",
                    "<<dmypy-goto-definition>>",
                    None,
                ),
                (
                    "Inspect Attributes",
                    "<<dmypy-inspect-attrs>>",
                    None,
                ),
            ),
        )

    def __getattr__(self, attr_name: str) -> object:
        """Transform event async sync calls to sync wrappers."""
        if attr_name.endswith("_event"):
            as_async = f"{attr_name}_async"
            if hasattr(self, as_async):
                return self.get_async(as_async)
        return super().__getattribute__(attr_name)

    def update_task_status(self, ignore: set[str] | None = None) -> None:
        """Update async task statusbar entry."""
        display = ""
        if hasattr(self.triorun, "nursery"):
            child_tasks = self.triorun.nursery.child_tasks
            task_names = {task.name.rsplit(".", 1)[-1] for task in child_tasks}
            if ignore:
                task_names -= ignore
            if task_names:
                tasks = (
                    name.removesuffix("_async").removesuffix("_event")
                    for name in sorted(task_names)
                )
                tasks = (" ".join(name.split("_")).title() for name in tasks)
                plural = "s" if len(task_names) > 1 else ""
                display = f"Async task{plural}: {', '.join(tasks)}"
        self.editwin.status_bar.set_label("asyncstatus", display, side="right")

    def get_async(
        self,
        name: str,
    ) -> Callable[[Event[Misc]], str]:
        """Get sync callable to run async function."""
        async_function = getattr(self, name)

        # Type of decorated function contains type `Any`
        @wraps(async_function)
        async def task_status_wrap(event: Event[Misc]) -> None:  # type: ignore[misc]
            self.update_task_status()
            try:
                await async_function(event)
            finally:
                self.update_task_status({name})

        # Type of decorated function contains type `Any`
        @wraps(async_function)
        @utils.log_exceptions
        def call_trio(event: Event[Misc]) -> str:  # type: ignore[misc]
            self.triorun(partial(task_status_wrap, event))
            return "break"

        return call_trio

    @property
    def daemon_timeout(self) -> int:
        """Daemon timeout."""
        if self.timeout_mins == "None":
            return DAEMON_TIMEOUT_MIN * 60
        try:
            return max(
                DAEMON_TIMEOUT_MIN * 60,
                math.ceil(float(self.timeout_mins) * 60),
            )
        except ValueError:
            return DAEMON_TIMEOUT_MIN * 60

    @property
    def action_timeout(self) -> int | None:
        """Action timeout."""
        if self.action_max_sec == "None":
            return None
        try:
            return max(ACTION_TIMEOUT_MIN, int(self.action_max_sec))
        except ValueError:
            return None

    @property
    def flags(self) -> list[str]:
        """Mypy Daemon flags."""
        base = {
            "--hide-error-context",
            "--no-color-output",
            "--show-absolute-path",
            "--no-error-summary",
            "--soft-error-limit=-1",
            "--show-traceback",
            f'--cache-dir="{self.mypy_folder}"',
            # f'--log-file="{self.log_file}"',
            # "--cache-fine-grained",
        }
        if self.daemon_flags == "None":
            return list(base)
        extra = set()
        for arg in self.daemon_flags.split():
            value = arg.strip()
            if value:
                extra.add(value)
        return list(base | extra)

    @property
    def typecomment_only_current_file(self) -> bool:
        """Should only add type comments for currently open file?."""
        return True

    def add_type_comments_for_file(
        self,
        comments: list[utils.Comment],
    ) -> dict[str, list[int]]:
        """Add type comments for target files.

        Return list of lines were a comment was added.

        Changes are wrapped in an undo block.
        """
        # Split up comments by line in order
        line_data: dict[int, list[utils.Comment]] = {}
        for comment in comments:
            line_data.setdefault(comment.line, [])
            line_data[comment.line].append(comment)

        all_messages = []
        for line in sorted(line_data):
            messages = line_data[line]
            if not messages:
                continue
            all_messages.extend(messages)
            pointers = self.get_pointers(messages)
            if pointers is not None:
                all_messages.append(pointers)

        return self.add_comments(all_messages)

    def add_mypy_messages(
        self,
        start_line: int,
        mypy_output: str,
        only_filename: str | None = None,
        add_all_override: bool = False,
    ) -> dict[str, list[int]]:
        """Add mypy comments for target filename.

        Return list of lines where comments were added.

        Changes are wrapped in an undo block.
        """
        default_file = UNKNOWN_FILE
        if self.files.filename is not None:
            default_file = os.path.abspath(self.files.filename)

        if only_filename is not None:
            only_filename = os.path.abspath(only_filename)

        file_comments = parse_comments(
            mypy_output,
            default_file,
            start_line,
        )

        file_commented_lines: dict[str, list[int]] = {}

        to_comment = set(file_comments)

        if self.typecomment_only_current_file and not add_all_override:
            assert only_filename is not None
            to_comment = {only_filename}

            # Find first line in target file or use start_line
            if not file_comments.get(only_filename):
                other_files_comment_line = start_line
            else:
                other_files_comment_line = min(
                    comment.line for comment in file_comments[only_filename]
                )

            # Add comments about how other files have errors
            file_comments.setdefault(only_filename, [])
            for filename in file_comments:
                if filename == only_filename:
                    continue
                file_comments[only_filename].append(
                    utils.Comment(
                        file=only_filename,
                        line=other_files_comment_line,
                        contents=f"Another file has errors: {filename!r}",
                        column_end=0,
                    ),
                )

        for target_filename in to_comment:
            if target_filename not in file_comments:
                continue
            if target_filename == UNKNOWN_FILE:
                continue
            file_comment_lines = self.add_type_comments_for_file(
                file_comments[target_filename],
            )
            file_commented_lines.update(file_comment_lines)
        return file_commented_lines

    def add_extra_data(
        self,
        file: str,
        start_line: int,
        data: str,
        prefix: str = "",
    ) -> tuple[int, list[int]]:
        """Add extra data to file as a big block of comments.

        Returns
        -------
        Tuple of:
        - Number of lines attempted to add
        - List of line numbers added that were not already there
        otherwise empty because no content.

        Changes are wrapped in an undo block.

        """
        if not data:
            return 0, []
        lines = data.splitlines()
        if not lines:
            return 0, []
        lines[0] = f"{prefix}{lines[0]}"
        added = self.add_comment_block(file, start_line, lines)
        return len(lines), added

    def add_errors(
        self,
        file: str,
        start_line: int,
        errors: str,
    ) -> tuple[int, list[int]]:
        """Add error lines to file as a block of comments.

        Returns
        -------
        Tuple of:
        - Number of lines attempted to add
        - List of line numbers added that were not already there
        otherwise None because no content.

        """
        return self.add_extra_data(
            file,
            start_line,
            errors,
            prefix="Error running mypy: ",
        )

    async def ensure_daemon_running(self) -> bool:
        """Make sure daemon is running. Return False if cannot continue."""
        if await client.is_running(self.status_file):
            return True
        command = " ".join(
            x
            for x in [
                "dmypy",
                f'--status-file="{self.status_file}"',
                "start",
                f'--log-file="{self.log_file}"',
                (
                    f"--timeout={self.daemon_timeout}"
                    if self.daemon_timeout
                    else ""
                ),
                "--",
                *self.flags,
            ]
            if x
        )
        debug(f"{command = }")
        return await client.start(
            self.status_file,
            flags=self.flags,
            daemon_timeout=self.daemon_timeout,
            log_file=self.log_file,
        )

    async def shutdown_dmypy_daemon_event_async(
        self,
        event: Event[Misc] | None = None,
    ) -> str:
        """Shutdown dmypy daemon event handler."""
        # pylint: disable=unused-argument
        if not await client.is_running(self.status_file):
            self.text.bell()
            return "break"

        # Only stop if running
        command = f'dmypy --status-file="{self.status_file}" stop'
        debug(f"{command = }")
        response = await client.stop(self.status_file)
        debug(f"{response = }")
        if response.get("err") or response.get("error"):
            # Kill
            await client.kill(self.status_file)

        return "break"

    async def check(self, file: str) -> client.Response:
        """Perform dmypy check."""
        if self.should_restart_always != "False":
            await self.shutdown_dmypy_daemon_event_async()
        if not await self.ensure_daemon_running():
            return COULD_NOT_START_ERROR
        ##flags = self.flags
        ##flags += [file]
        ### debug(f"check {flags = }")
        ##command = " ".join(
        ##    x
        ##    for x in [
        ##        "dmypy",
        ##        f'--status-file="{self.status_file}"',
        ##        "run",
        ##        (
        ##            f"--timeout={self.action_timeout}"
        ##            if self.action_timeout
        ##            else ""
        ##        ),
        ##        f'--log-file="{self.log_file}"',
        ##        "--export-types",
        ##        f'"{file}"',
        ##        "--",
        ##        *self.flags,
        ##    ]
        ##    if x
        ##)
        ##debug(f"{command = }")
        ##return await client.run(
        ##    self.status_file,
        ##    flags=flags,
        ##    timeout=self.action_timeout,
        ##    daemon_timeout=self.daemon_timeout,
        ##    log_file=self.log_file,
        ##    export_types=True,
        ##)
        command = f"dmypy --status-file='{self.status_file}' check --export-types {file!r}"
        debug(f"{command = }")
        return await client.check(
            self.status_file,
            files=[file],
            timeout=self.action_timeout,
            export_types=True,
        )

    def get_suggestion_text(
        self,
        annotation: dict[str, object],
    ) -> tuple[str | None, int]:
        """Get suggestion text from annotation.

        Return None on error or no difference, text if different
        """
        # while annotation["line"] >= 0 and "def" not in self.get_line(
        #     annotation["line"],
        # ):
        #     annotation["line"] -= 1
        line = annotation["line"]
        assert isinstance(line, int)

        try:
            text, line_count = annotate.get_annotation(
                annotation,
                self.get_line,
            )
        except Exception as ex:
            ex_text, ex_traceback = sys.exc_info()[1:]
            traceback.print_exception(
                None,  # Ignored since python 3.5
                value=ex_text,
                tb=ex_traceback,
                limit=None,
                chain=True,
            )
            indent = utils.get_line_indent(self.get_line(line))
            return (
                self.get_comment_line(
                    indent,
                    f"Error generating suggestion: {ex}",
                ),
                1,
            )

        select_start = f"{line}.0"
        line_end = line + line_count
        select_end = f"{line_end}.0"

        if text == self.text.get(select_start, select_end)[:-1]:
            return None, line_count
        return text, line_count

    def get_response_errors(self, response: client.Response) -> str | None:
        """Return errors from response if they exist else None."""
        errors = ""
        if response.get("error"):
            errors += response["error"]
        if response.get("err"):
            if errors:
                errors += "\n\n"
            errors += response["err"]
        if response.get("stderr"):
            if errors:
                errors += "\n\n"
            errors += f"stderr:\n{response['stderr']}"
        if "out" not in response and not errors:
            errors += "No response from dmypy daemon."

        return errors if errors else None

    async def suggest(self, file: str, line: int) -> None:
        """Perform dmypy suggest."""
        if await self.ensure_daemon_running():
            function = f"{file}:{line}"

            command = " ".join(
                (
                    "dmypy",
                    f'--status-file="{self.status_file}"',
                    "suggest",
                    f'"{function}"',
                ),
            )
            debug(f"{command = }")

            response = await client.suggest(
                self.status_file,
                function=function,
                do_json=True,
                timeout=self.action_timeout,
            )
        else:
            response = COULD_NOT_START_ERROR
        debug(f"suggest {response = }")

        if errors := self.get_response_errors(response):
            # Display errors
            # self.editwin.getlineno()
            self.add_errors(file, line, errors)
            self.text.bell()
            return

        annotations = orjson.loads(response["out"])

        line = annotations[0]["line"]

        samples: dict[int, list[str]] = {}
        line_count = 0
        for annotation in annotations:
            count = annotation["samples"]
            text, suggest_lines = self.get_suggestion_text(annotation)
            if text is None:
                continue
            if count not in samples:
                samples[count] = []
            samples[count].append(text)
            line_count += suggest_lines

        order = sorted(samples, reverse=True)
        lines = []
        for count in order:
            for sample in samples[count]:
                if sample not in lines:
                    lines.append(sample)

        replace = self.suggest_replace == "True"

        if len(lines) == 1:
            text = lines[0]
            if "Error generating suggestion: " in text:
                replace = False
        else:
            text = "\n".join(lines)
            replace = False

        select_start = f"{line}.0"
        line_end = line + line_count
        select_end = f"{line_end}.0"

        if not text or text == self.text.get(select_start, select_end)[:-1]:
            # Bell to let user know happened, just nothing to do
            self.editwin.gotoline(line)
            self.text.bell()
            return

        if not replace and "Error generating suggestion: " not in text:
            text = "\n".join(f"##{line}" for line in text.splitlines())
        text += "\n"

        with utils.undo_block(self.undo):
            if replace:
                self.text.delete(select_start, select_end)

            self.text.insert(select_start, text, ())

        self.editwin.gotoline(line)
        self.text.bell()

    def initial(self) -> tuple[str | None, str | None]:
        """Do common initial setup. Return error or none, file.

        Reload configuration, make sure file is saved,
        and make sure mypy is installed
        """
        # Reload configuration
        self.reload()

        # Get file we are checking
        raw_filename: str | None = self.files.filename
        if raw_filename is None:
            return "break", None
        file: str = os.path.abspath(raw_filename)

        # Make sure file is saved.
        if not self.files.get_saved():
            if not utils.ask_save_dialog(self.text):
                # If not ok to save, do not run. Would break file.
                self.text.bell()
                return "break", file
            # Otherwise, we are clear to save
            self.files.save(None)
            if not self.files.get_saved():
                return "break", file

        # Everything worked
        return None, file

    async def suggest_signature_event_async(self, event: Event[Misc]) -> str:
        """Handle suggest signature event."""
        # pylint: disable=unused-argument
        init_return, file = self.initial()

        if init_return is not None:
            return init_return

        if file is None:
            return "break"

        # if client.REQUEST_LOCK.locked():
        #     # If already requesting something from daemon,
        #     # do not send any other requests.
        #     debug(client.REQUEST_LOCK.statistics())
        #     # Make bell sound so user knows this ran even though
        #     # nothing happened.
        #     self.text.bell()
        #     return "break"

        await self.suggest(file, self.editwin.getlineno())

        return "break"

    async def dmypy_inspect(
        self,
        location: str,
        show: Literal["type", "attrs", "definition"],
        include_span: bool,
    ) -> client.Response:
        """Return result from dmypy inspect."""
        if await self.ensure_daemon_running():
            command = " ".join(
                x
                for x in (
                    "dmypy",
                    f"--status-file='{self.status_file}'",
                    "inspect",
                    f"--show={show!r}",
                    ("--include-span" if include_span else ""),
                    "--force-reload",
                    f"{location!r}",
                )
                if x
            )
            debug(f"{command = }")

            return await client.inspect(
                self.status_file,
                location=location,
                show=show,
                include_span=include_span,
                force_reload=True,  # needed_save,
            )
        return COULD_NOT_START_ERROR

    async def dmypy_inspect_shared(
        self,
        show: Literal["type", "attrs", "definition"],
        include_span: bool,
    ) -> (
        tuple[Literal[False], str] | tuple[Literal[True], tuple[str, str, int]]
    ):
        """Shared dmypy inspect code, return either fail string or success (output, file, line_no)."""
        # needed_save = not self.files.get_saved()
        init_return, file = self.initial()

        if init_return is not None:
            return False, init_return

        if file is None:
            return False, "break"

        sel_start, sel_end = utils.get_selected_text_indexes(self.text)

        start_line, start_col = utils.get_line_col(sel_start)
        end_line, end_col = utils.get_line_col(sel_end)

        location = f"{file}:{start_line}:{start_col + 1}"
        if sel_start != sel_end:
            location += f":{end_line}:{end_col}"

        result = await self.dmypy_inspect(location, show, include_span)

        if errors := self.get_response_errors(result):
            # Display errors
            # self.editwin.getlineno()
            with utils.undo_block(self.undo):
                self.add_errors(file, start_line, errors)
            self.text.bell()
            return False, "break"

        output = result["out"]
        if result.get("status"):
            with utils.undo_block(self.undo):
                self.add_errors(file, start_line, output)
            return False, "break"

        return True, (output, file, start_line)

    async def dmypy_inspect_type_event_async(self, event: Event[Misc]) -> str:
        """Perform dmypy inspect type from right click menu."""
        success, maybe_response = await self.dmypy_inspect_shared(
            show="type",
            include_span=True,
        )
        if not success:
            value = maybe_response
            assert not isinstance(value, tuple)
            # Failed somehow
            return value

        response_tuple = maybe_response
        assert isinstance(response_tuple, tuple)
        response, file, start_line = response_tuple

        comments = parse_type_inspect(response, file, start_line)

        for index, comment in enumerate(tuple(comments)):
            pointer = self.get_pointers([comment])
            if not pointer:
                continue
            comments[index] = pointer.replace_content(
                f"{pointer.contents} - {comment.contents}",
            )
        self.add_comments(comments)

        return "break"

    async def dmypy_goto_definition_event_async(
        self,
        event: Event[Misc],
    ) -> str:
        """Perform dmypy inspect definition from right click menu."""
        success, maybe_response = await self.dmypy_inspect_shared(
            show="definition",
            include_span=False,
        )
        if not success:
            value = maybe_response
            assert not isinstance(value, tuple)
            # Failed somehow
            return value

        response_tuple = maybe_response
        assert isinstance(response_tuple, tuple)
        raw_locations, file, start_line = response_tuple

        # Just read first one
        raw_location = raw_locations.splitlines()[0]

        debug(f"goto definition {raw_location = }")
        location, _function = raw_location.rsplit(":", 1)

        position = utils.FilePosition.parse(location)
        position = position.delta_column()

        # Try to get editor window of file path
        editor_window = self.flist.open(position.path)

        # On failure do a bell
        if editor_window is None:
            self.text.bell()
            return "break"

        # Show selection in file
        utils.show_hit(editor_window.text, *position.as_select(), tag="sel")

        return "break"

    async def dmypy_inspect_attrs_event_async(
        self,
        event: Event[Misc],
    ) -> str:
        """Perform dmypy inspect attributes from right click menu."""
        success, maybe_response = await self.dmypy_inspect_shared(
            show="attrs",
            include_span=True,
        )
        if not success:
            value = maybe_response
            assert not isinstance(value, tuple)
            # Failed somehow
            return value

        response_tuple = maybe_response
        assert isinstance(response_tuple, tuple)
        response, file, start_line = response_tuple

        comments: list[utils.Comment] = []

        for comment in parse_type_inspect(response, file, start_line):
            pointer = self.get_pointers([comment])
            contents = comment.contents
            data = orjson.loads(contents)
            lines = []
            for type_, attributes in data.items():
                lines.append(
                    utils.Comment(
                        file=file,
                        line=start_line + 1,
                        contents=f"type {type_}:",
                    ),
                )
                for attr in attributes:
                    lines.append(
                        utils.Comment(
                            file=file,
                            line=start_line + 1,
                            contents=f"    {attr}",
                        ),
                    )

            if pointer:
                comments.append(pointer)
            comments.extend(lines)
        self.add_comments(comments)

        return "break"

    def type_check_add_response_comments(
        self,
        response: client.Response,
        file: str,
    ) -> None:
        """Add all the comments (error and regular) from dmypy response."""
        debug(f"type check {response = }")

        if response.get("out"):
            # Add code comments
            self.add_mypy_messages(
                self.editwin.getlineno(),
                response["out"],
                file,
            )
        if response.get("error"):
            self.add_errors(file, self.editwin.getlineno(), response["error"])
        if response.get("err"):
            # Add mypy run errors
            self.add_errors(file, self.editwin.getlineno(), response["err"])
        if response.get("stdout"):
            self.add_extra_data(
                file,
                self.editwin.getlineno(),
                response["stdout"],
                prefix="dmypy run stdout: ",
            )
        if response.get("stderr"):
            self.add_extra_data(
                file,
                self.editwin.getlineno(),
                response["stderr"],
                prefix="dmypy run stderr: ",
            )

        # Make bell sound so user knows we are done,
        # as it freezes a bit while mypy looks at the file
        self.text.bell()

    async def type_check_event_async(self, event: Event[Misc]) -> str:
        """Perform a mypy check and add comments."""
        init_return, file = self.initial()

        if init_return is not None:
            return init_return

        if file is None:
            return "break"

        if client.REQUEST_LOCK.locked():
            # If already requesting something from daemon,
            # do not send any other requests.
            debug(client.REQUEST_LOCK.statistics())
            # Make bell sound so user knows this ran even though
            # nothing happened.
            self.text.bell()
            return "break"

        # Run mypy on open file
        response = await self.check(file)

        self.type_check_add_response_comments(response, file)
        return "break"

    @utils.log_exceptions
    def remove_type_comments_event(self, _event: Event[Misc]) -> str:
        """Remove selected extension comments."""
        self.remove_selected_extension_comments()
        return "break"

    @utils.log_exceptions
    def remove_all_type_comments(self, _event: Event[Misc]) -> str:
        """Remove all extension comments."""
        self.remove_all_extension_comments()
        return "break"

    @utils.log_exceptions
    def find_next_type_comment_event(self, _event: Event[Misc]) -> str:
        """Find next extension comment by hacking the search dialog engine."""
        # Reload configuration
        self.reload()

        # Find comment
        self.find_next_extension_comment(self.search_wrap == "True")

        return "break"

    def unregister_async_events(self) -> None:
        """Unregister asynchronous event handlers."""
        for bind_name in self.bind_defaults:
            attr_name = bind_name.replace("-", "_") + "_event_async"
            if hasattr(self, attr_name):
                self.text.event_delete(f"<<{bind_name}>>")

    @utils.log_exceptions
    def on_reload(self) -> None:
        """Extension cleanup before IDLE window closes."""
        # Wrapped in try except so failure doesn't cause zombie windows.
        with contextlib.suppress(AttributeError):
            del self.triorun
        try:
            self.unregister_async_events()
        except Exception as exc:
            utils.extension_log_exception(exc)



"""TKTrio - Run trio on top of Tkinter."""

# Programmed by CoolCat467

from __future__ import annotations

# TKTrio - Run Trio on top of Tkinter.
# Copyright (C) 2023-2025  CoolCat467
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__title__ = "TKTrio"
__author__ = "CoolCat467"
__license__ = "GNU General Public License Version 3"
__version__ = "0.1.0"

import contextlib
import queue
import sys
import threading
import tkinter as tk
import weakref
from enum import IntEnum, auto
from functools import partial, wraps
from tkinter import messagebox
from traceback import format_exception
from typing import TYPE_CHECKING, Any, TypeGuard

from idlemypyextension import utils
from idlemypyextension.moduleguard import guard_imports

with guard_imports({"trio", "exceptiongroup"}):
    import trio

    if sys.version_info < (3, 11):  # pragma: nocover
        from exceptiongroup import ExceptionGroup

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from outcome import Outcome
    from typing_extensions import Self, TypeVarTuple, Unpack

    PosArgT = TypeVarTuple("PosArgT")


def debug(message: object) -> None:  # pragma: nocover
    """Print debug message."""
    # TODO: Censor username/user files
    print(f"\n[{__title__}] DEBUG: {message}")


def check_main_thread() -> bool:
    """Return if this function is being run from main thread."""
    return threading.current_thread() is threading.main_thread()


def install_protocol_override(
    root: tk.Wm,
    override: Callable[
        [
            str | None,
            Callable[[], None] | None,
            Callable[[str | None, Callable[[], None] | None], str],
        ],
        str,
    ],
) -> None:
    """Install protocol override.

    Basically, whenever root.protocol is called, instead it calls
    override with the name and function to bind and also passes the
    original root.protocol function
    """
    uninstall_protocol_override(root)
    original = root.protocol

    @wraps(original)
    def new_protocol(
        name: str | None = None,
        func: Callable[[], None] | None = None,
    ) -> str:
        return override(name, func, original)  # type: ignore[arg-type]

    root.protocol = new_protocol  # type: ignore[assignment]


def uninstall_protocol_override(root: tk.Wm) -> None:
    """Uninstall protocol override if it exists.

    If root.protocol has the __wrapped__ attribute, reset it
    to what it was originally
    """
    if hasattr(root.protocol, "__wrapped__"):  # pragma: nocover
        root.protocol = root.protocol.__wrapped__  # type: ignore


class RunStatus(IntEnum):
    """Enum for trio run status.

    NO_TASK is when there is no task and not running
    TRIO_RUNNING is when there is task and running
    TRIO_RUNNING_CANCELED is when there is task and running but in the process
    of canceling task
    """

    TRIO_RUNNING_CANCELED = auto()
    TRIO_RUNNING_CANCELING = auto()
    TRIO_RUNNING = auto()
    TRIO_STARTING = auto()
    NO_TASK = auto()


def evil_does_trio_have_runner() -> bool:  # pragma: nocover
    """Evil function to see if trio has a runner."""
    core = getattr(trio, "_core", None)
    if core is None:
        return False
    run = getattr(core, "_run", None)
    if run is None:
        return False
    global_run_context = getattr(run, "GLOBAL_RUN_CONTEXT", None)
    if global_run_context is None:
        return False
    return hasattr(global_run_context, "runner")


def is_tk_wm_and_misc_subclass(
    val: tk.Toplevel | tk.Tk | object,
) -> TypeGuard[tk.Toplevel | tk.Tk]:
    """Return if value is an instance of tk.Toplevel."""
    return isinstance(val, tk.Toplevel | tk.Tk)


class ThreadRunner:
    """Thread runner."""

    __slots__ = ("check_period", "run_queue", "running")

    def __init__(self, check_period: int = 30) -> None:
        """Initialize thread runner."""
        self.running = False
        self.check_period = check_period

        self.run_queue: queue.Queue[
            tuple[
                Callable[[], Any],
                queue.Queue[
                    tuple[
                        bool,
                        Exception | None,
                    ]
                ],
            ]
        ] = queue.Queue(1)

    def check_events(self, tk: tk.Toplevel | tk.Tk) -> None:
        """Check for events to process."""
        not_empty = False
        while self.running:
            try:
                function, response_queue = self.run_queue.get_nowait()
            except queue.Empty:
                break
            else:
                not_empty = True
            try:
                function()
            except Exception as exc:
                response_queue.put((True, exc))
            else:
                response_queue.put((False, None))
        # Schedule to check again. If we just processed an event, check
        # immediately; if we didn't, check later.
        if not_empty:
            tk.after_idle(self.check_events, tk)
        elif self.running:
            tk.after(self.check_period, self.check_events, tk)

    def start_run(self, tk: tk.Toplevel | tk.Tk) -> None:
        """Start checking for events in this thread."""
        if self.running:
            raise RuntimeError("Already running from somewhere else!")
        self.running = True
        tk.after_idle(self.check_events, tk)

    def stop_run(self) -> None:
        """Stop running."""
        self.running = False

    def get_running(self) -> bool:
        """Return if is running."""
        return self.running

    def __call__(self, function: Callable[[], Any]) -> None:
        """Call a function using run in different thread."""
        if not self.running:
            raise RuntimeError("Not running")
        response_queue: queue.Queue[
            tuple[
                bool,
                Exception | None,
            ]
        ] = queue.Queue(1)
        self.run_queue.put(
            (
                function,
                response_queue,
            ),
            True,
            2,
        )
        is_exception, maybe_exc = response_queue.get(True, None)
        if is_exception:
            assert maybe_exc is not None
            raise maybe_exc


class TkTrioRunner:
    """Tk Trio Runner - Run Trio on top of Tkinter's main loop."""

    __slots__ = (
        "__weakref__",
        "call_tk_close",
        "installed_proto_override",
        "nursery",
        "received_loop_close_request",
        "root",
        "run_status",
        "thread_runner",
    )

    def __new__(
        cls,
        root: tk.Toplevel | tk.Tk,
        # hold_global_object: object | None,
        *args: Any,
        **kwargs: Any,
    ) -> Self:
        """Either return new instance or get existing runner from root."""
        if not is_tk_wm_and_misc_subclass(root):
            raise ValueError("Must be subclass of both tk.Misc and tk.Wm")
        # if hold_global_object is None:
        hold_global_object = root
        ref = getattr(hold_global_object, "__trio__", None)

        if ref is not None:
            instance = ref()
            if instance is not None:
                print(
                    f"[{__name__}]: {cls.__name__}: Loaded instance from hold_global_object",
                )
                if TYPE_CHECKING:
                    assert isinstance(instance, cls)
                assert instance.__class__.__name__ == cls.__name__
                return instance
        return super().__new__(cls)

    def __init__(
        self,
        root: tk.Toplevel | tk.Tk,
        # hold_global_object: object | None = None,
        restore_close: Callable[[], Any] | None = None,
    ) -> None:
        """Initialize trio runner."""
        # if hold_global_object is None:
        hold_global_object = root
        if (
            hasattr(hold_global_object, "__trio__")
            and getattr(hold_global_object, "__trio__", lambda: None)()
            is not None
        ):
            return
        self.root = root
        self.call_tk_close = restore_close or root.destroy

        self.nursery: trio.Nursery
        self.run_status = RunStatus.NO_TASK
        self.received_loop_close_request = False
        self.installed_proto_override = False

        with contextlib.suppress(AttributeError):
            hold_global_object.__trio__ = weakref.ref(self)  # type: ignore[union-attr]

        self.thread_runner = ThreadRunner()

    def schedule_task_threadsafe(
        self,
        function: Callable[[], object],
    ) -> None:
        """Schedule task in Tkinter's event loop."""
        try:
            self.thread_runner(function)
        except Exception as exc:
            debug(f"Exception scheduling task {function = }")
            # probably "main thread is not in main loop" error
            # mtTkinter is supposed to fix this sort of issue
            utils.extension_log_exception(exc)

            self.cancel_current_task()

    def schedule_task_not_threadsafe(
        self,
        function: Callable[[Unpack[PosArgT]], object],
        *args: Unpack[PosArgT],
    ) -> None:
        """Schedule task in Tkinter's event loop."""
        try:
            self.root.after_idle(function, *args)
        except Exception as exc:
            debug(f"Exception scheduling task {function = }")
            utils.extension_log_exception(exc)

            self.cancel_current_task()

    def cancel_current_task(self) -> None:
        """Cancel current task if one is running."""
        if self.run_status == RunStatus.NO_TASK:
            # No need to reschedule, is already closed.
            return

        if self.run_status == RunStatus.TRIO_STARTING:
            # Reschedule close for later, in process of starting.
            self.schedule_task_not_threadsafe(self.cancel_current_task)
            return

        if self.run_status == RunStatus.TRIO_RUNNING_CANCELING:
            # Already scheduled to cancel
            return

        self.run_status = RunStatus.TRIO_RUNNING_CANCELING
        try:
            self.nursery.cancel_scope.cancel()
        except RuntimeError as exc:
            # probably "must be called from async context" error
            # because the exception that triggered this was from
            # a start group tick failing because of start_soon
            # not running from main thread because thread lock shenanigans

            # Stop thread runner
            self.thread_runner.stop_run()

            utils.extension_log_exception(exc)

            # We can't even show an error properly because of the same
            # issue!
            try:
                self.show_irrecoverable_error_stopping(
                    "".join(format_exception(exc)),
                )
            except RuntimeError as exc:
                utils.extension_log_exception(exc)
        else:
            self.run_status = RunStatus.TRIO_RUNNING_CANCELED

    def _done_callback(self, outcome: Outcome[None]) -> None:
        """Handle when trio is done running."""
        # Stop thread runner
        self.thread_runner.stop_run()
        assert self.run_status in {
            RunStatus.TRIO_RUNNING_CANCELED,
            RunStatus.TRIO_RUNNING,
        }
        self.run_status = RunStatus.NO_TASK
        del self.nursery
        try:
            outcome.unwrap()
        except ExceptionGroup as exc:
            utils.extension_log_exception(exc)

    def _start_async_task(
        self,
        function: Callable[[], Awaitable[Any]],
    ) -> None:
        """Run async task in new nursery."""
        if evil_does_trio_have_runner():
            self.show_warning_trio_already_running()
            return

        @trio.lowlevel.disable_ki_protection
        async def run_nursery() -> None:
            """Run nursery."""
            assert self.run_status == RunStatus.TRIO_STARTING
            async with trio.open_nursery(
                strict_exception_groups=True,
            ) as nursery:
                self.nursery = nursery
                self.run_status = RunStatus.TRIO_RUNNING
                self.nursery.start_soon(function)
            return

        if self.run_status != RunStatus.NO_TASK:
            raise RuntimeError(
                "Cannot run more than one trio instance at once.",
            )

        if not check_main_thread():
            raise RuntimeError("Need to start from main thread.")

        # Start running thread runner from main thread
        if not self.thread_runner.get_running():
            self.thread_runner.start_run(self.root)

        self.run_status = RunStatus.TRIO_STARTING

        trio.lowlevel.start_guest_run(
            run_nursery,
            done_callback=self._done_callback,
            run_sync_soon_threadsafe=self.schedule_task_threadsafe,
            run_sync_soon_not_threadsafe=self.schedule_task_not_threadsafe,
            host_uses_signal_set_wakeup_fd=False,
            restrict_keyboard_interrupt_to_checkpoints=True,
            strict_exception_groups=True,
        )

    def get_del_window_proto(
        self,
        new_protocol: Callable[[], None] | None,
    ) -> Callable[[], None]:
        """Create new WM_DELETE_WINDOW protocol to shut down trio properly."""

        def shutdown_then_call() -> None:
            # If this is first time, withdraw root
            if not self.received_loop_close_request:
                self.root.withdraw()
            self.received_loop_close_request = True
            # If a task is still running
            if self.run_status == RunStatus.TRIO_RUNNING_CANCELING:
                # Error happened and cancel failed
                # Hopefully will be ok because closing host loop
                self.run_status = RunStatus.NO_TASK
                debug(
                    "Changing run status Canceling -> No Task, error happened during canceling.",
                )
            if self.run_status == RunStatus.TRIO_RUNNING:
                # Cancel it if not already canceled
                self.cancel_current_task()
            if self.run_status != RunStatus.NO_TASK:
                # Rerun this function again in the future until no task running.
                self.schedule_task_not_threadsafe(shutdown_then_call)
                return None
            # No more tasks
            # Make sure to uninstall override or IDLE gets mad
            uninstall_protocol_override(self.root)
            self.installed_proto_override = False
            # If close function exists, call it.
            if new_protocol is not None:
                return new_protocol()
            return None

        return shutdown_then_call

    def host_loop_close_override(
        self,
        name: str | None,
        new_protocol: Callable[[], None] | None,
        original_bind_proto: Callable[
            [str | None, Callable[[], None] | None],
            str,
        ],
    ) -> str:
        """Override for new protocols on tkinter root.

        Catch new protocols for `WM_DELETE_WINDOW` and replace them
        with our own that will cancel trio run and make sure it closes, and
        then call the new protocol
        """
        if name != "WM_DELETE_WINDOW":
            # Not important to us
            return original_bind_proto(name, new_protocol)
        return original_bind_proto(
            name,
            self.get_del_window_proto(new_protocol),
        )

    def show_warning_trio_already_running(self) -> None:
        """Show warning that trio is already running."""
        messagebox.showerror(
            title="Error: Trio is already running",
            message="Trio is already running from somewhere else, "
            "please try again later.",
            parent=self.root,
        )

    def show_irrecoverable_error_stopping(
        self,
        extra_message: str = "",
    ) -> None:
        """Show warning that previous trio run needs to stop."""
        messagebox.showwarning(
            title="Irrecoverable Error While Stopping",
            message=(
                "Encountered an irrecoverable error stopping a previous "
                "Trio run. Please restart IDLE and report this on Github at "
                "https://github.com/CoolCat467/idlemypyextension/issues"
                f"{extra_message}"
            ),
            parent=self.root,
        )

    def no_start_trio_is_stopping(self) -> None:
        """Show warning that previous trio run needs to stop."""
        messagebox.showwarning(
            title="Warning: Trio is stopping a previous run",
            message="Trio is in the process of stopping, "
            "please try again later.",
            parent=self.root,
        )

    def __call__(self, function: Callable[[], Awaitable[Any]]) -> None:
        """Schedule async function for the future."""
        # If host loop requested to close, do not run any more tasks.
        if self.received_loop_close_request:
            return None

        if self.run_status == RunStatus.TRIO_RUNNING_CANCELING:
            # Task status is in a limbo state where it tried to
            # cancel the nursery but something went wrong.
            # This should not happen with mtTkinter's fixes, but
            # it might still be a possibility.
            self.show_irrecoverable_error_stopping()
            return None

        if self.run_status == RunStatus.TRIO_RUNNING_CANCELED:
            ### Task run is in the process of stopping
            ### self.no_start_trio_is_stopping()
            # Reschedule starting task
            self.schedule_task_not_threadsafe(self.__call__, function)
            return None

        if self.run_status == RunStatus.TRIO_STARTING:
            # Reschedule starting task
            self.schedule_task_not_threadsafe(self.__call__, function)
            return None

        # If there is a task running
        if self.run_status == RunStatus.TRIO_RUNNING:
            self.nursery.start_soon(function)
            return None

        if self.run_status != RunStatus.NO_TASK:
            raise RuntimeError(f"Invalid run status {self.run_status!r}")

        # If we have not installed the protocol override,
        if not self.installed_proto_override:
            # Install it
            install_protocol_override(self.root, self.host_loop_close_override)
            self.installed_proto_override = True
            # Make sure new handler is now applied
            self.root.protocol("WM_DELETE_WINDOW", self.call_tk_close)

        # Start running new task
        return self._start_async_task(function)


def run() -> None:
    """Run test of module."""

    # A tiny Trio program
    async def trio_test_program(run: int) -> str:
        for i in range(5):
            print(f"Hello from Trio! ({i}) ({run = })")
            # This is inside Trio, so we have to use Trio APIs
            try:
                await trio.sleep(1)
            except trio.Cancelled:
                print(f"Run {run} canceled.")
                raise
        return "trio done!"

    root = tk.Tk()

    def trigger_trio_runs() -> None:
        trio_run = TkTrioRunner(root, None)
        trio_run(partial(trio_test_program, 0))
        trio_run_2 = TkTrioRunner(root, None)
        trio_run_2(partial(trio_test_program, 1))
        trio_run_2(partial(trio_test_program, 2))
        trio_run(partial(trio_test_program, 3))

    # Wait a tiny bit so main loop is running from main thread
    root.after(10, trigger_trio_runs)
    print("synchronous after trio run start")
    # If testing without IDLE, uncomment the line below:
    # root.mainloop()


if __name__ == "__main__":
    print(f"{__title__}\nProgrammed by {__author__}.\n")
    run()



from __future__ import annotations

from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest
import trio

from idlemypyextension.tktrio import RunStatus, TkTrioRunner

if TYPE_CHECKING:
    import tkinter as tk
    from collections.abc import Callable, Generator

    from typing_extensions import TypeVarTuple, Unpack

    PosArgT = TypeVarTuple("PosArgT")


@pytest.fixture(autouse=True)
def mock_extension_log() -> Generator[MagicMock, None, None]:
    """Fixture to override extension_log with an empty function."""
    with patch(
        "idlemypyextension.utils.extension_log",
        return_value=None,
    ) as mock_log:
        yield mock_log


class FakeTK:
    """Fake tkinter root."""

    __slots__ = ("__dict__", "should_stop", "tasks")

    def __init__(self) -> None:
        """Initialize self."""
        self.tasks: list[Callable[..., object]] = []
        self.should_stop = False

    def after_idle(
        self,
        function: Callable[[Unpack[PosArgT]], object],
        *args: Unpack[PosArgT],
    ) -> None:
        """Add function to run queue."""
        self.tasks.append(partial(function, *args))

    def after(
        self,
        _delay: int,
        function: Callable[[Unpack[PosArgT]], object],
        *args: Unpack[PosArgT],
    ) -> None:
        """Add function to run queue."""
        self.tasks.append(partial(function, *args))

    def update(self) -> None:
        """Run one task from queue."""
        if self.should_stop:
            raise RuntimeError(f"{self.should_stop = }")
        if self.tasks:
            self.tasks.pop(0)()

    def destroy(self) -> None:
        """Mark update to fail if called again."""
        self.should_stop = True

    def withdraw(self) -> None:
        """Fake hide window."""

    def protocol(
        self,
        name: str | None,
        func: Callable[[], None] | None = None,
    ) -> None:
        """Fake protocol."""


@pytest.fixture
def mock_tk() -> FakeTK:
    """Fixture to create a mock tkinter root."""
    return FakeTK()


@pytest.fixture
def trio_runner(mock_tk: tk.Tk) -> TkTrioRunner:
    """Fixture to create a TkTrioRunner instance."""
    with patch(
        "idlemypyextension.tktrio.is_tk_wm_and_misc_subclass",
        return_value=True,
    ):
        return TkTrioRunner(mock_tk, None)


def test_initialization(trio_runner: TkTrioRunner) -> None:
    """Test initialization of TkTrioRunner."""
    assert trio_runner.run_status == RunStatus.NO_TASK
    assert trio_runner.root is not None


def test_new_gives_copy(mock_tk: tk.Tk) -> None:
    with patch(
        "idlemypyextension.tktrio.is_tk_wm_and_misc_subclass",
        return_value=True,
    ):
        runner = TkTrioRunner(mock_tk, None)
        runner2 = TkTrioRunner(mock_tk, None)
        assert runner is runner2


def test_invalid_initialization() -> None:
    """Test initialization with invalid root."""
    with pytest.raises(
        ValueError,
        match=r"^Must be subclass of both tk\.Misc and tk\.Wm$",
    ):
        TkTrioRunner(None, None)  # type: ignore[arg-type]


def test_schedule_task_not_threadsafe(trio_runner: TkTrioRunner) -> None:
    """Test scheduling a task in the Tkinter event loop."""
    mock_function = MagicMock()
    trio_runner.schedule_task_not_threadsafe(mock_function)

    # Process the scheduled tasks
    trio_runner.root.update()

    mock_function.assert_called_once()


def test_cancel_current_task(trio_runner: TkTrioRunner) -> None:
    """Test canceling the current task."""
    assert trio_runner.run_status != RunStatus.TRIO_RUNNING

    async def test() -> None:
        while True:
            await trio.lowlevel.checkpoint()

    trio_runner(test)

    while trio_runner.run_status == RunStatus.TRIO_STARTING:
        trio_runner.root.update()

    nursery = trio_runner.nursery

    trio_runner.cancel_current_task()

    assert trio_runner.run_status == RunStatus.TRIO_RUNNING_CANCELED

    while str(trio_runner.run_status) != str(RunStatus.NO_TASK):
        trio_runner.root.update()

    assert nursery.cancel_scope.cancel_called

    trio_runner.cancel_current_task()


def test_get_del_window_proto(trio_runner: TkTrioRunner) -> None:
    """Test the WM_DELETE_WINDOW protocol."""
    new_protocol = MagicMock()
    shutdown_function = trio_runner.get_del_window_proto(new_protocol)

    # Call the shutdown function
    shutdown_function()

    # Check if the new protocol was called
    new_protocol.assert_called_once()


def test_show_warning_trio_already_running(trio_runner: TkTrioRunner) -> None:
    """Test showing warning when Trio is already running."""
    with patch("tkinter.messagebox.showerror") as mock_showerror:
        trio_runner.show_warning_trio_already_running()
        mock_showerror.assert_called_once_with(
            title="Error: Trio is already running",
            message="Trio is already running from somewhere else, please try again later.",
            parent=trio_runner.root,
        )


def test_no_start_trio_is_stopping(trio_runner: TkTrioRunner) -> None:
    """Test showing warning when Trio is stopping."""
    with patch("tkinter.messagebox.showwarning") as mock_showwarning:
        trio_runner.no_start_trio_is_stopping()
        mock_showwarning.assert_called_once_with(
            title="Warning: Trio is stopping a previous run",
            message="Trio is in the process of stopping, please try again later.",
            parent=trio_runner.root,
        )



"""Type Check IDLE Extension."""

# Programmed by CoolCat467

from __future__ import annotations

# Copyright (C) 2023-2025  CoolCat467
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__title__ = "idlemypyextension"
__author__ = "CoolCat467"
__license__ = "GNU General Public License Version 3"
__version__ = "1.4.0"


import argparse
import idlelib.pyshell as pyshell
import sys
import tkinter as tk

from idlemypyextension import utils
from idlemypyextension.extension import idlemypyextension as idlemypyextension


def check_installed() -> bool:
    """Make sure extension installed. Return True if installed correctly."""
    return utils.check_installed(__title__, __version__, idlemypyextension)


def run(args: list[str]) -> int:
    """Entry point."""
    parser = argparse.ArgumentParser(
        prog=__title__,
        description="Mypy Daemon IDLE Integration Extension.",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"{__title__} v{__version__}",
    )
    parser.add_argument(
        "infile",
        nargs="?",
        type=argparse.FileType("r"),
        default=sys.stdin,
    )

    namespace = parser.parse_args(args)

    if not args:
        if check_installed():
            return 0
        return 1

    original_capture_warnings = pyshell.capture_warnings

    def capture_warnings(enable: bool) -> None:
        """Only enable capturing warnings."""
        if enable:
            original_capture_warnings(True)

    # Start IDLE but make sure mainloop does not start so we can use it
    # ourselves
    class ExitException(BaseException):
        """Exit Exception."""

    def fake_mainloop(tk_root: tk.Tk) -> None:
        """Fake Tkinter main loop. Crash instantly."""
        raise ExitException()

    try:
        with utils.temporary_overwrite(tk.Tk, "mainloop", fake_mainloop):
            with utils.temporary_overwrite(
                sys,
                "argv",
                [pyshell.__file__, "-e"],  # open in editor mode
            ):
                pyshell.main()
    except ExitException:
        pass  # expected to happen
    else:
        # Something went wrong, did not get exit exception
        pyshell.root.destroy()
        pyshell.capture_warnings(False)
        return 1

    # Get file list and Tk Root from pyshell globals
    flist = pyshell.flist

    pyshell_window = {v: k for k, v in flist.inversedict.items()}.get(None)

    if pyshell_window is None:
        # Pyshell not created properly, exit.
        pyshell.root.destroy()
        pyshell.capture_warnings(False)
        return 1
    assert isinstance(pyshell_window, pyshell.PyShellEditorWindow)

    if not pyshell_window.extensions:
        # No extensions loaded, close.
        pyshell.root.destroy()
        pyshell.capture_warnings(False)
        return 1

    extension = pyshell_window.extensions.get(__title__)

    if extension is None:
        # Extension failed to load somehow, close
        pyshell.root.destroy()
        pyshell.capture_warnings(False)
        return 1
    assert isinstance(extension, idlemypyextension)

    # Add all comments to all files.
    with namespace.infile as fp:
        extension.add_mypy_messages(0, fp.read(), add_all_override=True)
    # Close original blank window
    del extension
    pyshell_window.close()

    # Run IDLE's main loop
    while flist.inversedict:  # keep IDLE running while files are open.
        pyshell.root.mainloop()
    pyshell.root.destroy()
    pyshell.capture_warnings(False)

    return 0


def cli_run() -> None:
    """Command line interface entry point."""
    sys.exit(run(sys.argv[1:]))


utils.set_title(__title__)
idlemypyextension.reload()


if __name__ == "__main__":
    print(f"{__title__} v{__version__}\nProgrammed by {__author__}.\n")
    cli_run()
